"""Sync tests for REST position liquidation endpoints."""

from typing import List
from pydantic import BaseModel


def test_rest_list_position_liquidations(rc):
    """Test listing position liquidations."""
    liquidations = rc.list_position_liquidations()
    assert isinstance(liquidations, List)
    assert all(isinstance(liq, BaseModel) for liq in liquidations)



import logging
from typing import Union, Dict, Any
from ethereal.models.config import BaseConfig


def get_logger(name):
    """Get a configured logger instance.

    Args:
        name (str): Name for the logger

    Returns:
        Logger: Configured logging instance
    """
    logger = logging.getLogger(name)

    # Check if the logger already has handlers
    if not logger.handlers and logger.propagate:
        logger.setLevel(logging.INFO)
        return logger

    # If no handlers are set, configure a default handler
    if not logger.handlers and not logger.propagate:
        logger.setLevel(logging.INFO)
        handler = logging.StreamHandler()
        formatter = logging.Formatter(
            "%(asctime)s - %(name)s - %(levelname)s - %(message)s", "%Y-%m-%d %H:%M:%S"
        )
        handler.setFormatter(formatter)
        logger.addHandler(handler)

    return logger


class BaseClient:
    """Base client with common functionality.

    Args:
        config (Union[Dict[str, Any], BaseConfig]): Base configuration
    """

    def __init__(self, config: Union[Dict[str, Any], BaseConfig]):
        self.config = BaseConfig.model_validate(config)
        self._setup_logging()

    def _setup_logging(self):
        """Set up logging for the client."""
        self.logger = get_logger(self.__class__.__name__)
        if self.config.verbose:
            self.logger.setLevel(logging.DEBUG)
        pass



#!/usr/bin/env python3
"""
Generates a synchronous REST client from the async client.

This script introspects the AsyncRESTClient and automatically generates
a synchronous wrapper. The wrapper runs a dedicated asyncio event loop in a
background thread and schedules coroutines via run_coroutine_threadsafe.
This avoids conflicts in environments with a running loop (e.g. marimo/Jupyter).
"""

import inspect
import re
from pathlib import Path
from typing import ForwardRef, get_args, get_origin, get_type_hints, Union
from ethereal.async_rest_client import AsyncRESTClient

SKIP_IMPORT_NAMES = {
    "Any",
    "Dict",
    "List",
    "Optional",
    "Tuple",
    "Union",
    "UUID",
    "Decimal",
}


def format_type_hint(type_obj) -> str:
    """Format a type hint for the generated code."""
    if type_obj is None or type_obj is type(None):
        return "None"

    # Handle None type
    if type_obj is type(None):
        return "None"

    # Basic types
    if type_obj in (int, str, bool, float):
        return type_obj.__name__

    # Handle forward references and string annotations early
    if isinstance(type_obj, ForwardRef):
        return type_obj.__forward_arg__
    if isinstance(type_obj, str):
        return _clean_type_string(type_obj)

    # Handle typing module types
    if hasattr(type_obj, "__module__"):
        if type_obj.__module__ == "typing":
            origin = get_origin(type_obj)
            args = get_args(type_obj)

            if origin is None:
                # Handle typing objects without origin (e.g., Any)
                return str(type_obj).replace("typing.", "")
            elif origin is list:
                if args:
                    return f"List[{format_type_hint(args[0])}]"
                return "List"
            elif origin is dict:
                if len(args) == 2:
                    return f"Dict[{format_type_hint(args[0])}, {format_type_hint(args[1])}]"
                return "Dict"
            elif origin is tuple:
                if args:
                    formatted_args = ", ".join(format_type_hint(arg) for arg in args)
                    return f"Tuple[{formatted_args}]"
                return "Tuple"
            elif origin is Union:
                formatted_args = ", ".join(format_type_hint(arg) for arg in args)
                return f"Union[{formatted_args}]"
            else:
                # Generic typing construct
                return str(type_obj).replace("typing.", "")

        # Handle model types
        elif "ethereal.models" in type_obj.__module__:
            return type_obj.__name__

        # Handle other module types
        elif type_obj.__module__ not in ("builtins", "__main__"):
            return type_obj.__name__

    # Handle class types
    if hasattr(type_obj, "__name__"):
        return type_obj.__name__

    # Fallback to string representation
    type_str = _clean_type_string(str(type_obj))
    return type_str


def _clean_type_string(type_str: str) -> str:
    """Normalize a type string by stripping module prefixes and ForwardRefs."""
    # Clean up common patterns
    type_str = re.sub(r"<class '([^']+)'>", r"\1", type_str)
    type_str = type_str.replace("typing.", "")
    type_str = type_str.replace("ethereal.models.rest.", "")
    type_str = type_str.replace("NoneType", "None")
    type_str = re.sub(r"""ForwardRef\((?:'|")([^'"]+)(?:'|")\)""", r"\1", type_str)
    return type_str.replace("'", "").replace('"', "")


def get_method_signature(method):
    """Get the signature and call arguments for a method."""
    try:
        sig = inspect.signature(method)
        # Try to get type hints, fall back to raw annotations if forward references fail
        module = inspect.getmodule(method)
        globalns = vars(module) if module else {}
        try:
            type_hints = get_type_hints(method, globalns=globalns, localns=globalns)
        except (NameError, AttributeError, TypeError):
            type_hints = getattr(method, "__annotations__", {})
    except (TypeError, ValueError) as e:
        print(
            f"Warning: Could not introspect {getattr(method, '__name__', 'unknown')}: {e}"
        )
        return "(self, *args, **kwargs)", "*args, **kwargs", "Any"

    params = []
    call_args = []

    for param_name, param in sig.parameters.items():
        if param_name == "self":
            params.append(param_name)
            continue

        # Get type hint
        type_hint = type_hints.get(param_name, param.annotation)
        if type_hint is inspect.Parameter.empty:
            type_hint = "Any"
        formatted_type = format_type_hint(type_hint)

        # Clean up type hint formatting for known issues
        if "pydantic.main.BaseModel" in formatted_type:
            formatted_type = formatted_type.replace(
                "pydantic.main.BaseModel", "BaseModel"
            )

        # Build parameter string
        if param.kind == inspect.Parameter.VAR_POSITIONAL:
            params.append(f"*{param_name}")
            call_args.append(f"*{param_name}")
        elif param.kind == inspect.Parameter.VAR_KEYWORD:
            params.append(f"**{param_name}")
            call_args.append(f"**{param_name}")
        else:
            if param.default is inspect.Parameter.empty:
                param_str = f"{param_name}: {formatted_type}"
            else:
                default_val = repr(param.default)
                # Handle special cases for default values
                if default_val.startswith("<") and default_val.endswith(">"):
                    default_val = "..."
                param_str = f"{param_name}: {formatted_type} = {default_val}"
            params.append(param_str)
            call_args.append(param_name)

    # Get return type
    return_annotation = sig.return_annotation
    if return_annotation is inspect.Signature.empty:
        return_type = type_hints.get("return", "Any")
    else:
        return_type = return_annotation
    return_type_str = format_type_hint(return_type)

    signature = f"({', '.join(params)}) -> {return_type_str}"
    call_args_str = ", ".join(call_args)

    return signature, call_args_str, return_type_str


def generate_sync_client():
    """Generate a .pyi type stub for the sync REST client.

    - Introspects AsyncRESTClient to collect method signatures and types.
    - Dynamically derives DTO imports used in signatures to avoid drift.
    """

    # Template for the generated stub
    template = """# This file is auto-generated by scripts/generate_sync_client.py
# Type stub for the synchronous REST client.

from __future__ import annotations

from typing import Any, Dict, List, Optional, Tuple, Type, Union
from pydantic import BaseModel
from decimal import Decimal
from ethereal.models.config import RESTConfig
from ethereal.chain_client import ChainClient
from uuid import UUID
{dto_imports}


class RESTClient:
    def __init__(self, config: Union[Dict[str, Any], RESTConfig] = ...) -> None: ...
    def close(self) -> None: ...
    def __enter__(self) -> RESTClient: ...
    def __exit__(self, exc_type, exc_val, exc_tb) -> None: ...

{methods}

    # Exposed internals
    @property
    def chain(self) -> Optional[ChainClient]: ...
    @property
    def logger(self) -> Any: ...
    @property
    def provider(self) -> Any: ...
    @property
    def private_key(self) -> Optional[str]: ...
    @property
    def _models(self) -> Any: ...
"""

    # Generate methods
    methods = []
    dto_names = set()

    # Get all async methods from AsyncRESTClient
    for method_name, member in inspect.getmembers(AsyncRESTClient):
        # Skip private methods, special methods, and non-coroutine functions
        if (
            method_name.startswith("_")
            or method_name in ["create", "close"]
            or not inspect.iscoroutinefunction(member)
        ):
            continue

        # Generate method wrapper
        sig, call_args, return_type = get_method_signature(member)
        # Collect DTOs referenced in type hints dynamically
        try:
            hints = get_type_hints(member)
        except Exception:
            hints = getattr(member, "__annotations__", {})
        for t in list(hints.values()):
            origin = get_origin(t)
            args = get_args(t) if origin else ()
            to_check = args or (t,)
            for cand in to_check:
                if isinstance(cand, str):
                    for match in re.findall(r"[A-Z][A-Za-z0-9_]+", cand):
                        if match not in SKIP_IMPORT_NAMES:
                            dto_names.add(match)
                    continue

                if isinstance(cand, ForwardRef):
                    forward_name = cand.__forward_arg__
                    if (
                        forward_name
                        and forward_name[0].isupper()
                        and forward_name not in SKIP_IMPORT_NAMES
                    ):
                        dto_names.add(forward_name)
                    continue

                mod = getattr(cand, "__module__", "")
                dto_name = getattr(cand, "__name__", None)
                if (
                    mod.startswith("ethereal.models")
                    and dto_name
                    and dto_name[0].isupper()
                    and dto_name not in SKIP_IMPORT_NAMES
                ):
                    dto_names.add(dto_name)

        method_code = f"""
    def {method_name}{sig}: ..."""
        methods.append(method_code)

    # Add _get_pages method
    methods.append("""
    def _get_pages(
        self,
        endpoint: str,
        request_model: Type[BaseModel],
        response_model: Type[BaseModel],
        paginate: bool = False,
        **kwargs
    ) -> Any:
        ...""")

    # Build dynamic DTO imports block
    if dto_names:
        names_sorted = ",\n    ".join(sorted(dto_names))
        dto_imports = f"from ethereal.models.rest import (\n    {names_sorted},\n)"
    else:
        dto_imports = "# No DTOs referenced"

    # Generate the final code
    generated_code = template.replace("{methods}", "\n".join(methods)).replace(
        "{dto_imports}", dto_imports
    )

    # Determine project root (two levels up from this script file)
    project_root = Path(__file__).resolve().parents[1]
    # Write to stub file
    output_path = project_root / "ethereal" / "rest_client.pyi"
    with open(output_path, "w") as f:
        f.write(generated_code)

    print(f"Successfully generated {output_path}")


if __name__ == "__main__":
    generate_sync_client()



from typing import Union, Dict, Any, Callable, Optional

from ethereal.models.config import WSConfig
from ethereal.ws.ws_base import WSBase


class WSClient(WSBase):
    """Ethereal websocket client.

    Args:
        config (Union[Dict[str, Any], WSConfig]): Configuration dictionary or WSConfig object.
            Required fields include:
            - base_url (str): Base URL for websocket requests
            Optional fields include:
            - verbose (bool): Enables debug logging, defaults to False
    """

    def __init__(self, config: Union[Dict[str, Any], WSConfig]):
        super().__init__(config)

    def subscribe(
        self,
        stream_type: str,
        product_id: Optional[str] = None,
        subaccount_id: Optional[str] = None,
        callback: Optional[Callable[[Dict[str, Any]], None]] = None,
        namespace: Optional[str] = "/v1/stream",
    ) -> Dict[str, Any]:
        """Subscribe to a specific stream.

        Args:
            stream_type (str): Type of stream to subscribe to
            product_id (Optional[str]): Product ID to subscribe to
            subaccount_id (Optional[str]): Subaccount ID, optional
            callback (Optional[Callable]): Callback function to handle incoming messages

        Returns:
            Dict[str, Any]: Subscription response
        """
        subscription_data = {"type": stream_type}

        if subaccount_id:
            subscription_data["subaccountId"] = subaccount_id
        if product_id:
            subscription_data["productId"] = product_id

        # Register callback if provided
        if callback:
            event_key = stream_type
            if event_key not in self.callbacks:
                self.callbacks[event_key] = []

            self.callbacks[event_key].append(callback)

        # Send subscription request
        return self._emit("subscribe", subscription_data, namespace=namespace)

    def unsubscribe(
        self,
        stream_type: str,
        product_id: Optional[str] = None,
        subaccount_id: Optional[str] = None,
        namespace: Optional[str] = "/v1/stream",
    ) -> Dict[str, Any]:
        """Unsubscribe from a specific stream.

        Args:
            stream_type (str): Type of stream to unsubscribe from
            product_id (str): Product ID to unsubscribe from
            subaccount_id (Optional[str]): Subaccount ID, optional

        Returns:
            Dict[str, Any]: Unsubscription response
        """
        unsubscription_data = {"type": stream_type}

        if subaccount_id:
            unsubscription_data["subaccountId"] = subaccount_id
        if product_id:
            unsubscription_data["productId"] = product_id

        # Send unsubscription request
        return self._emit("unsubscribe", unsubscription_data, namespace=namespace)



"""Pure sync tests for REST info endpoints that require an account."""

from decimal import Decimal
from typing import List
from pydantic import BaseModel


def test_rest_info_subaccount_balances(rc, sid):
    balances = rc.get_subaccount_balances(subaccount_id=sid)
    assert isinstance(balances, List)
    assert all(isinstance(sb, BaseModel) for sb in balances)


def test_rest_info_orders(rc, sid):
    orders = rc.list_orders(subaccount_id=sid)
    assert isinstance(orders, List)
    assert all(isinstance(o, BaseModel) for o in orders)


def test_rest_info_fills(rc, sid):
    fills = rc.list_fills(subaccount_id=sid)
    assert isinstance(fills, List)
    assert all(isinstance(f, BaseModel) for f in fills)


def test_rest_info_fills_paginated(rc, sid):
    fills = rc._get_pages(
        endpoint="order/fill",
        request_model=rc._models.V1OrderFillGetParametersQuery,
        response_model=rc._models.PageOfOrderFillDtos,
        subaccount_id=sid,
        limit=200,
        paginate=True,
    )
    assert isinstance(fills, List)
    assert all(isinstance(f, BaseModel) for f in fills)


def test_rest_info_trades(rc, sid):
    products = rc.list_products()
    params = {"product_id": products[0].id, "order": "desc", "limit": 100}
    trades = rc.list_trades(**params)
    assert isinstance(trades, List)
    assert all(isinstance(t, BaseModel) for t in trades)


def test_rest_info_positions(rc, sid):
    positions = rc.list_positions(subaccount_id=sid)
    assert isinstance(positions, List)
    assert all(isinstance(p, BaseModel) for p in positions)


def test_rest_info_maintenance_margin(rc, sid):
    mm = rc.get_maintenance_margin(subaccount_id=sid)
    assert isinstance(mm, Decimal)

    # specify product id
    pid = rc.products_by_ticker["BTCUSD"].id
    btc_mm = rc.get_maintenance_margin(subaccount_id=sid, product_ids=[pid])
    assert isinstance(btc_mm, Decimal)

    # specify products
    products = [
        rc.products_by_ticker["BTCUSD"],
        rc.products_by_ticker["ETHUSD"],
    ]
    eth_btc_mm = rc.get_maintenance_margin(subaccount_id=sid, products=products)
    assert isinstance(eth_btc_mm, Decimal)

    # specify product dicts
    product_dicts = [
        {
            "id": rc.products_by_ticker["BTCUSD"].id,
            "max_leverage": 5,
            "taker_fee": "0.001",
        },
        {
            "id": rc.products_by_ticker["ETHUSD"].id,
            "max_leverage": 5,
            "taker_fee": "0.001",
        },
    ]
    eth_btc_mm = rc.get_maintenance_margin(subaccount_id=sid, products=product_dicts)
    assert isinstance(eth_btc_mm, Decimal)



import time
from decimal import Decimal
from typing import List, Optional
from enum import Enum
from uuid import UUID

from ethereal.constants import API_PREFIX
from ethereal.models.rest import (
    TokenDto,
    WithdrawDto,
    TransferDto,
    InitiateWithdrawDto,
)
from ethereal.rest.util import ensure_bytes32_hex, generate_nonce


class LayerZeroEndpointId(Enum):
    NONE = 0
    ETHEREAL_TESTNET = 40422
    ARBITRUM_SEPOLIA = 40231


async def list_tokens(
    self,
    **kwargs,
) -> List[TokenDto]:
    """Lists all supported tokens.

    Other Parameters:
        deposit_enabled (bool, optional): Filter tokens with deposits enabled. Optional.
        withdraw_enabled (bool, optional): Filter tokens with withdrawals enabled. Optional.
        order (str, optional): Sort order, 'asc' or 'desc'. Optional.
        limit (float, optional): Maximum number of results to return. Optional.
        cursor (str, optional): Pagination cursor for fetching the next page. Optional.
        order_by (str, optional): Field to order by (e.g., 'createdAt'). Optional.
        **kwargs: Additional request parameters accepted by the API. Optional.

    Returns:
        List[TokenDto]: Token metadata.
    """
    res = await self.get_validated(
        url_path=f"{API_PREFIX}/token",
        request_model=self._models.V1TokenGetParametersQuery,
        response_model=self._models.PageOfTokensDtos,
        **kwargs,
    )
    data = [
        self._models.TokenDto(**model.model_dump(by_alias=True)) for model in res.data
    ]
    return data


async def get_token(
    self,
    id: UUID,
    **kwargs,
) -> TokenDto:
    """Gets a specific token by ID.

    Args:
        id (str): UUID for the token. Required.

    Other Parameters:
        **kwargs: Additional request parameters accepted by the API. Optional.

    Returns:
        TokenDto: Token metadata.
    """
    endpoint = f"{API_PREFIX}/token/{id}"
    res = await self.get(endpoint, **kwargs)
    return self._models.TokenDto(**res)


async def list_token_withdraws(
    self,
    **kwargs,
) -> List[WithdrawDto]:
    """Lists token withdrawals for a subaccount.

    Args:
        subaccount_id (str): UUID of the subaccount. Required.

    Other Parameters:
        is_active (bool, optional): Filter active withdrawals. Optional.
        order (str, optional): Sort order, 'asc' or 'desc'. Optional.
        limit (float, optional): Maximum number of results to return. Optional.
        cursor (str, optional): Pagination cursor for fetching the next page. Optional.
        order_by (str, optional): Field to order by (e.g., 'createdAt'). Optional.
        **kwargs: Additional request parameters accepted by the API. Optional.

    Returns:
        List[WithdrawDto]: Withdrawal records for the subaccount.
    """
    res = await self.get_validated(
        url_path=f"{API_PREFIX}/token/withdraw",
        request_model=self._models.V1TokenWithdrawGetParametersQuery,
        response_model=self._models.PageOfWithdrawDtos,
        **kwargs,
    )
    data = [
        self._models.WithdrawDto(**model.model_dump(by_alias=True))
        for model in res.data
    ]
    return data


async def list_token_transfers(
    self,
    **kwargs,
) -> List[TransferDto]:
    """Lists token transfers for a subaccount.

    Args:
        subaccount_id (str): UUID of the subaccount. Required.

    Other Parameters:
        statuses (List[str], optional): Filter by transfer status (e.g., 'COMPLETED'). Optional.
        types (List[str], optional): Filter by transfer type (e.g., 'WITHDRAW'). Optional.
        created_after (float, optional): Filter transfers created after this timestamp. Optional.
        created_before (float, optional): Filter transfers created before this timestamp. Optional.
        order (str, optional): Sort order, 'asc' or 'desc'. Optional.
        limit (float, optional): Maximum number of results to return. Optional.
        cursor (str, optional): Pagination cursor for fetching the next page. Optional.
        order_by (str, optional): Field to order by (e.g., 'createdAt'). Optional.
        **kwargs: Additional request parameters accepted by the API. Optional.

    Returns:
        List[TransferDto]: Transfer records for the subaccount.
    """
    res = await self.get_validated(
        url_path=f"{API_PREFIX}/token/transfer",
        request_model=self._models.V1TokenTransferGetParametersQuery,
        response_model=self._models.PageOfTransfersDtos,
        **kwargs,
    )
    data = [
        self._models.TransferDto(**model.model_dump(by_alias=True))
        for model in res.data
    ]
    return data


async def prepare_withdraw_token(
    self,
    subaccount: str,
    token: str,
    amount: int,
    account: str,
    destination_address: str,
    destination_endpoint: int,
    include_signature: bool = False,
    nonce: Optional[str] = None,
    signed_at: Optional[int] = None,
) -> InitiateWithdrawDto:
    """Prepares a token withdrawal request, optionally including a signature.

    Args:
        subaccount (str): Hex-encoded subaccount name. Required.
        token (str): Token contract address or identifier. Required.
        amount (int): Amount to withdraw (token base units). Required.
        account (str): Recipient address. Required.
        destination_address (str): LayerZero destination address (hex string or bytes; automatically left-padded to bytes32).
        destination_endpoint (int): LayerZero destination endpoint ID.
        include_signature (bool): If True, sign the payload immediately. Optional.
        nonce (str, optional): Custom nonce for signing. Defaults to a generated nonce.
        signed_at (int, optional): Seconds since epoch for the signature timestamp.

    Returns:
        InitiateWithdrawDto: Prepared (and optionally signed) withdrawal payload.
    """
    nonce = nonce or generate_nonce()
    signed_at = signed_at or int(time.time())
    lz_destination_id = self._models.LzDestinationEid(destination_endpoint)

    data = {
        "account": account,
        "subaccount": subaccount,
        "token": token,
        "amount": amount,
        "nonce": nonce,
        "signedAt": signed_at,
        "lzDestinationAddress": ensure_bytes32_hex(destination_address),
        "lzDestinationEid": lz_destination_id,
    }
    data_model = self._models.InitiateWithdrawDtoData.model_validate(data)
    dto = self._models.InitiateWithdrawDto.model_validate(
        {"data": data_model.model_dump(mode="json", by_alias=True), "signature": ""}
    )
    if include_signature:
        dto = await self.sign_withdraw_token(dto)
    return dto


async def sign_withdraw_token(
    self,
    withdraw_dto: InitiateWithdrawDto,
    private_key: Optional[str] = None,
) -> InitiateWithdrawDto:
    """Signs the token withdrawal payload using EIP-712.

    Args:
        withdraw_dto (InitiateWithdrawDto): Prepared withdrawal payload. Required.
        private_key (str, optional): Private key override. Defaults to the client's chain key.

    Returns:
        InitiateWithdrawDto: The same DTO with signature populated.

    Raises:
        ValueError: If no chain client or private key is available.
    """
    if not hasattr(self, "chain") or not self.chain:
        raise ValueError("No chain client available for signing")
    if not private_key and not self.chain.private_key:
        raise ValueError("No private key available for signing")
    elif not private_key:
        private_key = self.chain.private_key

    # Prepare the message for signing
    message = withdraw_dto.data.model_dump(by_alias=True, mode="json")
    message["amount"] = int(Decimal(str(message["amount"])) * Decimal(1e9))
    message["signedAt"] = int(message["signedAt"])

    destination_address = message["lzDestinationAddress"]
    destination_eid = int(message["lzDestinationEid"])

    message["destinationAddress"] = destination_address
    message["destinationEndpointId"] = destination_eid

    primary_type = "InitiateWithdraw"
    domain = self.rpc_config.domain.model_dump(mode="json", by_alias=True)
    types = self.chain.get_signature_types(self.rpc_config, primary_type)
    withdraw_dto.signature = self.chain.sign_message(
        private_key, domain, types, primary_type, message
    )
    return withdraw_dto


async def withdraw_token(
    self,
    dto: InitiateWithdrawDto,
    token_id: UUID,
    **kwargs,
) -> WithdrawDto:
    """Submits a prepared and signed token withdrawal request.

    Args:
        dto (InitiateWithdrawDto): Prepared and signed withdrawal payload. Required.
        token_id (str): Token ID for the withdraw route parameter. Required.

    Other Parameters:
        **kwargs: Additional request parameters accepted by the API. Optional.

    Returns:
        WithdrawDto: Withdrawal record created by the API.
    """
    endpoint = f"{API_PREFIX}/token/{token_id}/withdraw"
    res = await self.post(
        endpoint,
        data=dto.model_dump(mode="json", by_alias=True, exclude_none=True),
        **kwargs,
    )
    return self._models.WithdrawDto.model_validate(res)



from __future__ import annotations
import importlib
import warnings
from decimal import Decimal
from pydantic import BaseModel, AnyHttpUrl
from typing import (
    TYPE_CHECKING,
    Union,
    Dict,
    Any,
    Optional,
    Set,
    Type,
    List,
    Tuple,
)
from uuid import UUID
from ethereal.constants import API_PREFIX, ARCHIVE_NETWORK_URLS, NETWORK_URLS
from ethereal.rest.async_http_client import AsyncHTTPClient
from ethereal.chain_client import ChainClient
from ethereal.models.config import RESTConfig, ChainConfig

from ethereal.rest.funding import list_funding, get_projected_funding
from ethereal.rest.order import (
    get_order,
    list_fills,
    list_orders,
    list_trades,
    prepare_order,
    sign_order,
    submit_order,
    dry_run_order,
    prepare_cancel_order,
    sign_cancel_order,
    cancel_order,
)
from ethereal.rest.linked_signer import (
    get_signer,
    get_signer_quota,
    list_signers,
    prepare_linked_signer,
    sign_linked_signer,
    link_linked_signer,
    prepare_revoke_linked_signer,
    sign_revoke_linked_signer,
    revoke_linked_signer,
)
from ethereal.rest.position import (
    list_positions,
    get_position,
    list_position_liquidations,
)
from ethereal.rest.product import (
    get_market_liquidity,
    list_market_prices,
    list_products,
)
from ethereal.rest.rpc import get_rpc_config
from ethereal.rest.subaccount import (
    list_subaccounts,
    get_subaccount,
    get_subaccount_balances,
    get_subaccount_balance_history,
    get_subaccount_unrealized_pnl_history,
    get_subaccount_volume_history,
    get_subaccount_funding_history,
)
from ethereal.rest.token import (
    get_token,
    list_token_withdraws,
    list_tokens,
    list_token_transfers,
    prepare_withdraw_token,
    sign_withdraw_token,
    withdraw_token,
)

_MODEL_PATHS = {
    "default": "ethereal.models.rest",
    "mainnet": "ethereal.models.mainnet.rest",
    "testnet": "ethereal.models.testnet.rest",
    "devnet": "ethereal.models.devnet.rest",
}

if TYPE_CHECKING:
    from ethereal.models.rest import *  # noqa: F403


class PaginatedResponse(BaseModel):
    data: Any
    has_next: bool
    next_cursor: Optional[str]


class AsyncRESTClient(AsyncHTTPClient):
    """Asynchronous REST client for the Ethereal API.

    Notes for maintainers:
    - This client composes endpoint functions from the `ethereal.rest.*` modules
      by assigning them as attributes on the class (see below). Each function
      expects `self` to provide `get`, `post`, and `get_validated` from
      AsyncHTTPClient and to expose `_models` for the active network.
    - Network-specific models are accessed via `self._models` which is set based
      on the configured network. This avoids global mutation and is predictable.
    - Use `AsyncRESTClient.create(...)` to ensure async initialization (RPC
      config, optional chain client) happens before use. Remember to `await
      client.close()` to release the underlying `httpx.AsyncClient`.
    - We intentionally avoid “async properties”. For convenience methods that
      derive indices (e.g., products by ticker/id), use explicit async methods
      like `get_products_by_ticker()` or `get_products_by_id()` that fetch fresh
      data each call to keep behavior lightweight and predictable.
    """

    list_funding = list_funding
    get_projected_funding = get_projected_funding
    get_order = get_order
    list_fills = list_fills
    list_orders = list_orders
    list_trades = list_trades
    prepare_order = prepare_order
    sign_order = sign_order
    submit_order = submit_order
    dry_run_order = dry_run_order
    prepare_cancel_order = prepare_cancel_order
    sign_cancel_order = sign_cancel_order
    cancel_order = cancel_order
    get_signer = get_signer
    get_signer_quota = get_signer_quota
    list_signers = list_signers
    prepare_linked_signer = prepare_linked_signer
    sign_linked_signer = sign_linked_signer
    link_linked_signer = link_linked_signer
    prepare_revoke_linked_signer = prepare_revoke_linked_signer
    sign_revoke_linked_signer = sign_revoke_linked_signer
    revoke_linked_signer = revoke_linked_signer
    list_positions = list_positions
    get_position = get_position
    list_position_liquidations = list_position_liquidations
    get_market_liquidity = get_market_liquidity
    list_market_prices = list_market_prices
    list_products = list_products
    get_rpc_config = get_rpc_config
    list_subaccounts = list_subaccounts
    get_subaccount = get_subaccount
    get_subaccount_balances = get_subaccount_balances
    get_subaccount_balance_history = get_subaccount_balance_history
    get_subaccount_unrealized_pnl_history = get_subaccount_unrealized_pnl_history
    get_subaccount_volume_history = get_subaccount_volume_history
    get_subaccount_funding_history = get_subaccount_funding_history
    get_token = get_token
    list_token_withdraws = list_token_withdraws
    list_tokens = list_tokens
    list_token_transfers = list_token_transfers
    prepare_withdraw_token = prepare_withdraw_token
    sign_withdraw_token = sign_withdraw_token
    withdraw_token = withdraw_token

    def __init__(self, config: Union[Dict[str, Any], RESTConfig] = {}):
        self.config = RESTConfig.model_validate(config)
        network = self.config.network or "testnet"

        if not self.config.base_url:
            self.config.base_url = AnyHttpUrl(
                NETWORK_URLS.get(network, NETWORK_URLS["testnet"])
            )

        if not self.config.archive_base_url:
            self.config.archive_base_url = AnyHttpUrl(
                ARCHIVE_NETWORK_URLS.get(network, ARCHIVE_NETWORK_URLS["testnet"])
            )

        self._models = importlib.import_module(
            _MODEL_PATHS.get(network, _MODEL_PATHS["default"])  # type: ignore
        )
        super().__init__(self.config)

        self._archive_base_url = self.config.archive_base_url
        self.chain: Optional[ChainClient] = None
        self.rpc_config: Optional[RpcConfigDto] = None
        self.private_key: Optional[str] = None
        self.provider: Optional[Any] = None
        self.default_time_in_force = self.config.default_time_in_force
        self.default_post_only = self.config.default_post_only
        self._subaccounts: Optional[List[SubaccountDto]] = None
        self._products: Optional[List[ProductDto]] = None
        self._tokens: Optional[List[TokenDto]] = None
        self._products_by_ticker: Optional[Dict[str, ProductDto]] = None
        self._products_by_id: Optional[Dict[UUID, ProductDto]] = None

    @classmethod
    async def create(
        cls, config: Union[Dict[str, Any], RESTConfig] = {}
    ) -> AsyncRESTClient:
        """Factory method to create and asynchronously initialize the client.

        Args:
            config (Union[Dict[str, Any], RESTConfig], optional): Configuration dictionary or RESTConfig object. Optional fields include:
                private_key (str, optional): The private key.
                base_url (str, optional): Base URL for REST requests. Defaults to mainnet.
                timeout (int, optional): Timeout in seconds for REST requests.
                verbose (bool, optional): Enables debug logging. Defaults to False.
                rate_limit_headers (bool, optional): Enables rate limit headers. Defaults to False.
                chain_config (ChainConfig, optional): Chain configuration for signing transactions.

        Returns:
            AsyncRESTClient: Fully initialized async client instance.
        """
        client = cls(config)
        await client._async_init()
        return client

    async def _async_init(self):
        """Asynchronous initialization.

        Loads RPC configuration and (optionally) initializes the chain client if
        `chain_config` is provided. This split constructor pattern keeps
        `__init__` sync, while `create()` ensures the instance is fully ready.
        """
        self.rpc_config = await self.get_rpc_config()
        tokens = await self.list_tokens()
        if self.config.chain_config:
            self._init_chain_client(self.config.chain_config, self.rpc_config, tokens)
        self.private_key = self.chain.private_key if self.chain else None
        self.provider = self.chain.provider if self.chain else None
        # No eager cache priming; caches are populated on first access

    def _init_chain_client(
        self,
        config: Union[Dict[str, Any], ChainConfig],
        rpc_config: Optional[RpcConfigDto] = None,
        tokens: Optional[List[TokenDto]] = None,
    ):
        """Initialize the ChainClient for transaction signing.

        Args:
            config (Union[Dict[str, Any], ChainConfig]): The chain configuration.
            rpc_config (RpcConfigDto, optional): RPC configuration. Defaults to None.
            tokens (List[TokenDto], optional): List of token configurations. Defaults to None.
        """
        config = ChainConfig.model_validate(config)
        try:
            self.chain = ChainClient(config, rpc_config, tokens)
            self.logger.debug("Chain client initialized successfully")
        except Exception as e:
            self.logger.debug(f"Failed to initialize chain client: {e}")

    async def _get_pages(
        self,
        endpoint: str,
        request_model: Type[BaseModel],
        response_model: Type[BaseModel],
        paginate: bool = False,
        max_pages: int = 10,
        **kwargs,
    ) -> Any:
        """Make a GET request with validated parameters and response, handling pagination.

        Args:
            endpoint (str): API endpoint path (e.g. "order" will be appended to the base URL and prefix to form "/v1/order")
            request_model (Type[BaseModel]): Pydantic model for request validation
            response_model (Type[BaseModel]): Pydantic model for response validation
            paginate (bool, optional): If True, fetch all pages and return aggregated list. Defaults to False.
            max_pages (int, optional): Maximum number of pages to fetch when paginating. Defaults to 10.

        Other Parameters:
            **kwargs: Parameters to validate and include in the request

        Returns:
            Any: List of validated response objects. Single page if paginate=False, all pages if paginate=True.
        """
        first_page = await self.get_validated(
            url_path=f"{API_PREFIX}/{endpoint}",
            request_model=request_model,
            response_model=response_model,
            **kwargs,
        )
        # `get_validated` already returns an instance of `response_model` with
        # `.data`, `.has_next`, and `.next_cursor`.
        if not paginate:
            return list(first_page.data)

        all_items: List[Any] = list(first_page.data)
        current_cursor = getattr(first_page, "next_cursor", None)
        has_next = getattr(first_page, "has_next", False)

        page_count = 1
        while has_next and current_cursor and page_count <= max_pages:
            page = await self.get_validated(
                url_path=f"{API_PREFIX}/{endpoint}",
                request_model=request_model,
                response_model=response_model,
                cursor=current_cursor,
                **kwargs,
            )
            all_items.extend(page.data)
            has_next = getattr(page, "has_next", False)
            current_cursor = getattr(page, "next_cursor", None)
            page_count += 1

        return all_items

    async def subaccounts(self, refresh: bool = False) -> List[SubaccountDto]:
        """Get the list of subaccounts.

        Args:
            refresh (bool, optional): If True, bypass cache and fetch fresh data. Defaults to False.

        Returns:
            List[SubaccountDto]: List of subaccount objects for the connected wallet address.

        Raises:
            ValueError: If no chain client is configured or address is unavailable.
        """
        if not self.chain or not getattr(self.chain, "address", None):
            raise ValueError("Chain address is required to list subaccounts")
        if refresh or self._subaccounts is None:
            self._subaccounts = await self.list_subaccounts(
                sender=self.chain.address, order_by="createdAt", order="asc"
            )
        return self._subaccounts

    async def products(self, refresh: bool = False) -> List[ProductDto]:
        """Get the list of products.

        Args:
            refresh (bool, optional): If True, bypass cache and fetch fresh data. Defaults to False.

        Returns:
            List[ProductDto]: List of product objects.
        """
        if refresh or self._products is None:
            self._products = await self.list_products()
        return self._products

    async def tokens(self, refresh: bool = False) -> List[TokenDto]:
        """Get the list of tokens.

        Args:
            refresh (bool, optional): If True, bypass cache and fetch fresh data. Defaults to False.

        Returns:
            List[TokenDto]: List of token objects.
        """
        if refresh or self._tokens is None:
            self._tokens = await self.list_tokens()
        return self._tokens

    async def products_by_ticker(self, refresh: bool = False) -> Dict[str, ProductDto]:
        """Get the products indexed by ticker.

        Args:
            refresh (bool, optional): If True, bypass cache and fetch fresh data. Defaults to False.

        Returns:
            Dict[str, ProductDto]: Dictionary of products keyed by ticker.
        """
        if refresh or self._products_by_ticker is None:
            products = await self.products(refresh=refresh)
            self._products_by_ticker = {
                p.ticker: p for p in products if getattr(p, "ticker", None)
            }
        return self._products_by_ticker

    async def products_by_id(self, refresh: bool = False) -> Dict[UUID, ProductDto]:
        """Get the products indexed by ID.

        Args:
            refresh (bool, optional): If True, bypass cache and fetch fresh data. Defaults to False.

        Returns:
            Dict[UUID, ProductDto]: Dictionary of products keyed by ID.
        """
        if refresh or self._products_by_id is None:
            products = await self.products(refresh=refresh)
            self._products_by_id = {p.id: p for p in products}
        return self._products_by_id

    async def get_maintenance_margin(
        self,
        subaccount_id: UUID,
        positions: Optional[List[PositionDto] | List[Dict[str, Any]]] = None,
        products: Optional[List[ProductDto] | List[Dict[str, Any]]] = None,
        product_ids: Optional[List[UUID]] = None,
    ) -> Decimal:
        """Calculate the an account's maintenance margin for specified positions or products.

        Args:
            subaccount_id (str): Fetch positions for this subaccount when ``positions`` is not supplied.
            positions (List[PositionDto] | List[Dict[str, Any]], optional): Pre-fetched positions to use directly.
            products (List[ProductDto] | List[Dict[str, Any]], optional): Pre-fetched products used to filter
                the calculation.
            product_ids (List[str] | List[UUID], optional): Filters the calculation to these product IDs.

        Returns:
            Decimal: Total maintenance margin for the filtered positions.

        Raises:
            ValueError: If neither positions nor subaccount context is provided, or if any
                referenced product cannot be resolved.
        """

        class PartialPosition(BaseModel):
            product_id: UUID
            cost: str

        class PartialProduct(BaseModel):
            id: UUID
            max_leverage: float
            taker_fee: str

        if positions is None:
            positions = await self.list_positions(subaccount_id=subaccount_id)

        if products and product_ids:
            raise ValueError("Can only specify one of products and product_ids")

        if products is not None:
            valid_products = [
                PartialProduct(**p if isinstance(p, dict) else p.model_dump())
                for p in products
            ]
            products_by_id: Dict[UUID, PartialProduct] = {
                product.id: product for product in valid_products
            }
        else:
            raw_products_by_id = await self.products_by_id()
            products_by_id = {
                UUID(str(k)): PartialProduct(
                    **v if isinstance(v, dict) else v.model_dump()
                )
                for k, v in raw_products_by_id.items()
            }

        product_ids_set: Optional[Set[UUID]] = None
        if product_ids is not None:
            product_ids_set = set([UUID(str(pid)) for pid in product_ids])
            products_by_id_keys = set(products_by_id.keys())
            missing_products = product_ids_set - products_by_id_keys
            if missing_products:
                missing = ", ".join(sorted([str(p) for p in missing_products]))
                raise ValueError(f"Products not found for calculation: {missing}")

        valid_positions = [
            PartialPosition(**p if isinstance(p, dict) else p.model_dump())
            for p in positions
        ]
        positions_filtered = [
            position
            for position in valid_positions
            if product_ids_set is None or position.product_id in product_ids_set
        ]

        if not positions_filtered:
            return Decimal("0")

        total_mm = Decimal("0")
        for position in positions_filtered:
            product = products_by_id.get(position.product_id)
            if product is None:
                raise ValueError(
                    f"Product '{position.product_id}' not found for position '{position.product_id}'"
                )

            notional = abs(Decimal(position.cost))
            max_leverage = Decimal(str(product.max_leverage))
            taker_fee_rate = Decimal(product.taker_fee)

            mmr = Decimal("1") / (max_leverage * Decimal("2"))
            total_mm += notional * mmr
            total_mm += notional * taker_fee_rate

        return total_mm

    async def get_tokens(self) -> List[TokenDto]:
        """Return the latest list of tokens (no caching)."""
        return await self.list_tokens()

    async def _resolve_sender_and_subaccount(
        self, sender: Optional[str] = None, subaccount: Optional[str] = None
    ) -> Tuple[str, str]:
        if sender is None and self.chain:
            sender = self.chain.address
            warnings.warn(
                "Automatically resolving 'sender' parameter is deprecated and will be removed in a future version. "
                "Always explicitly pass 'sender' and 'subaccount' to avoid issues while using multiple account. "
                f"Using sender='{sender}'",
                DeprecationWarning,
                stacklevel=3,
            )
        if subaccount is None:
            subaccounts = await self.subaccounts()
            if not subaccounts:
                raise ValueError(
                    "No subaccounts found for this account. Please create a subaccount first."
                )
            self.logger.debug(
                f"First subaccount name: '{subaccounts[0].name}', id: '{subaccounts[0].id}'"
            )
            self.logger.debug(f"All subaccount names: {[s.name for s in subaccounts]}")
            subaccount = subaccounts[0].name
            warnings.warn(
                "Automatically resolving 'subaccount' parameter is deprecated and will be removed in a future version. "
                "Always explicitly pass 'subaccount' and 'subaccount' to avoid issues while using multiple account. "
                f"Using subaccount='{subaccount}'",
                DeprecationWarning,
                stacklevel=3,
            )
        return sender, subaccount

    async def create_order(
        self,
        order_type: str,
        quantity: float,
        side: int,
        price: Optional[float] = None,
        ticker: Optional[str] = None,
        product_id: Optional[UUID] = None,
        client_order_id: Optional[str] = None,
        sender: Optional[str] = None,
        subaccount: Optional[str] = None,
        time_in_force: Optional[str] = None,
        post_only: Optional[bool] = None,
        reduce_only: Optional[bool] = False,
        close: Optional[bool] = None,
        stop_price: Optional[float] = None,
        stop_type: Optional[int] = None,
        expires_at: Optional[int] = None,
        group_id: Optional[str] = None,
        group_contingency_type: Optional[int] = None,
        sign: bool = True,
        dry_run: bool = False,
        submit: bool = True,
    ) -> Union[SubmitOrderCreatedDto, DryRunOrderCreatedDto, SubmitOrderDto]:
        """Create and submit an order.

        Args:
            order_type (str): 'LIMIT' or 'MARKET'. Required.
            quantity (float): Order size. Required.
            side (int): 0 for buy, 1 for sell. Required.
            price (float, optional): Limit price for LIMIT orders.
            ticker (str, optional): Ticker of the product.
            product_id (str, optional): UUID of the product.
            client_order_id (str, optional): Subaccount-scoped client-generated id (UUID or <=32 alphanumeric).
            sender (str, optional): Address placing the order. Defaults to chain address.
            subaccount (str, optional): Hex-encoded subaccount name. Defaults to first subaccount.
            time_in_force (str, optional): For LIMIT orders (e.g., 'GTC', 'GTD'). Defaults to 'GTC'.
            post_only (bool, optional): For LIMIT orders; rejects if crossing. Defaults to False.
            reduce_only (bool, optional): If True, order only reduces position. Defaults to False.
            close (bool, optional): For MARKET orders; If True, closes the position.
            stop_price (float, optional): Stop trigger price.
            stop_type (int, optional): Stop type, either 0 (take-profit) or 1 (stop-loss), requires non-zero stopPrice.
            expires_at (int, optional): Expiry timestamp for GTD.
            group_id (str, optional): Group Id (UUID) for linking orders together in OCO/OTO relationships.
            group_contingency_type (int, optional): Contingency type for order groups: 0=OTO (Order-Triggers-Order), 1=OCO (One-Cancels-Other).
            sign (bool, optional): If True, sign the payload immediately. Defaults to True.
            dry_run (bool, optional): If True, validate without execution. Defaults to False.
            submit (bool, optional): If True, submit the order. Defaults to True.

        Returns:
            Union[SubmitOrderCreatedDto, DryRunOrderCreatedDto, SubmitOrderDto]: Created order response, dry-run validation result, or prepared order payload.

        Raises:
            ValueError: If neither product_id nor ticker is provided or if order type is invalid.
        """
        sender, subaccount = await self._resolve_sender_and_subaccount(
            sender, subaccount
        )

        if product_id is not None:
            products_by_id = await self.products_by_id()
            onchain_id = products_by_id[product_id].onchain_id
        elif ticker is not None:
            products_by_ticker = await self.products_by_ticker()
            onchain_id = products_by_ticker[ticker].onchain_id
        else:
            raise ValueError("Either product_id or ticker must be provided")

        order_params = {
            "sender": sender,
            "subaccount": subaccount,
            "side": side,
            "quantity": quantity,
            "onchain_id": onchain_id,
            "order_type": order_type,
            "client_order_id": client_order_id,
            "reduce_only": reduce_only,
            "close": close,
            "stop_price": stop_price,
            "stop_type": stop_type,
            "group_id": group_id,
            "group_contingency_type": group_contingency_type,
        }

        if order_type == "LIMIT":
            order_params.update(
                {
                    "price": price,
                    "time_in_force": time_in_force or self.default_time_in_force,
                    "post_only": post_only or self.default_post_only,
                    "expires_at": expires_at,
                }
            )
        elif order_type != "MARKET":
            raise ValueError("Invalid order type")

        order = await self.prepare_order(**order_params, include_signature=sign)
        if dry_run:
            return await self.dry_run_order(order)
        elif submit:
            return await self.submit_order(order)
        else:
            return order

    async def cancel_orders(
        self,
        order_ids: List[UUID],
        sender: Optional[str] = None,
        subaccount: Optional[str] = None,
        client_order_ids: List[str] = [],
        sign: bool = True,
        submit: bool = True,
        **kwargs,
    ) -> Union[List[CancelOrderResultDto], CancelOrderDto]:
        """Prepares and optionally submits a request to cancel multiple orders.

        Args:
            order_ids (List[str]): Order UUIDs to cancel. Required.
            sender (str, optional): Address initiating the cancellation. Defaults to chain address.
            subaccount (str, optional): Hex-encoded subaccount name. Defaults to first subaccount.
            client_order_ids (List[str], optional): Client-generated IDs to cancel. Defaults to empty list.
            sign (bool, optional): If True, sign the payload immediately. Defaults to True.
            submit (bool, optional): If True, submit the request to the API. Defaults to True.

        Other Parameters:
            **kwargs: Additional request parameters accepted by the API.

        Returns:
            Union[List[CancelOrderResultDto], CancelOrderDto]: Cancellation results per order id or prepared cancel payload.

        Raises:
            ValueError: If no order IDs or client order IDs provided for cancellation.
        """
        if len(order_ids) == 0 and len(client_order_ids) == 0:
            raise ValueError(
                "No order IDs or client order IDs provided for cancellation"
            )
        sender, subaccount = await self._resolve_sender_and_subaccount(
            sender, subaccount
        )
        try:
            prepared_cancel = await self.prepare_cancel_order(
                order_ids=order_ids,
                client_order_ids=client_order_ids,
                sender=sender,
                subaccount=subaccount,
                include_signature=sign,
                **kwargs,
            )
        except ValueError as e:
            self.logger.warning(f"Could not prepare/sign order cancellation: {e}")
            raise

        if not submit:
            return prepared_cancel

        result = await self.cancel_order(
            prepared_cancel,
            **kwargs,
        )
        return result

    async def cancel_all_orders(
        self,
        subaccount_id: UUID,
        product_ids: Optional[List[UUID]] = None,
        **kwargs,
    ) -> List[CancelOrderResultDto]:
        """Cancel all orders for a given subaccount.

        Args:
            subaccount_id (str): UUID of the subaccount. Required.
            product_ids (List[str], optional): Filter cancellation by product IDs.

        Other Parameters:
            **kwargs: Additional request parameters accepted by the API.

        Returns:
            List[CancelOrderResultDto]: Cancellation results per order id.

        Raises:
            ValueError: If no orders found to cancel or cancellation fails.
        """
        subaccount = await self.get_subaccount(id=subaccount_id)
        query_params = {
            "subaccount_id": subaccount_id,
            "isWorking": True,
            **kwargs,
        }
        if product_ids:
            query_params["product_ids"] = product_ids

        orders = await self._get_pages(
            endpoint="order",
            request_model=self._models.V1OrderGetParametersQuery,
            response_model=self._models.PageOfOrderDtos,
            paginate=True,
            **query_params,
        )
        order_ids = [order.id for order in orders]

        if len(order_ids) == 0:
            raise ValueError("No order IDs provided for cancellation")
        cancel_results = await self.cancel_orders(
            order_ids=order_ids,
            sender=subaccount.account,
            subaccount=subaccount.name,
            sign=True,
            submit=True,
        )
        if not isinstance(cancel_results, list):
            raise ValueError("Failed to cancel orders")
        return cancel_results

    async def replace_order(
        self,
        order: Optional[OrderDto] = None,
        order_id: Optional[UUID] = None,
        quantity: Optional[float] = None,
        price: Optional[float] = None,
        time_in_force: Optional[str] = None,
        post_only: Optional[bool] = None,
        reduce_only: Optional[bool] = False,
    ) -> Tuple[SubmitOrderCreatedDto, bool]:
        """Replace an existing order with new parameters.

        Args:
            order (OrderDto, optional): Existing order object to replace.
            order_id (str, optional): UUID of the order to replace.
            quantity (float, optional): New order size.
            price (float, optional): New limit price.
            time_in_force (str, optional): New time in force.
            post_only (bool, optional): New post-only flag.
            reduce_only (bool, optional): New reduce-only flag. Defaults to False.

        Returns:
            Tuple[SubmitOrderCreatedDto, bool]: Created order response and success flag.

        Raises:
            ValueError: If neither order nor order_id is provided, or both are provided.
        """
        if order is None and order_id is None:
            raise ValueError("Either order or order_id must be provided")
        elif order is not None and order_id is not None:
            raise ValueError("Only one of order or order_id must be provided")
        elif order is not None:
            old_order = order
        elif order_id is not None:
            old_order = await self.get_order(id=order_id)
        subaccount = await self.get_subaccount(id=old_order.subaccount_id)

        quantity = float(old_order.quantity) if quantity is None else quantity
        price = float(old_order.price) if price is None else price
        time_in_force = (
            old_order.time_in_force.value
            if time_in_force is None and old_order.time_in_force
            else time_in_force
        )
        post_only = old_order.post_only if post_only is None else post_only
        reduce_only = old_order.reduce_only if reduce_only is None else reduce_only

        cancel_result = await self.cancel_orders(
            order_ids=[old_order.id],
            sender=old_order.sender,
            subaccount=subaccount.name,
            sign=True,
            submit=True,
        )
        if not isinstance(cancel_result, list) or len(cancel_result) != 1:
            raise ValueError("Failed to cancel order")
        canceled_order = cancel_result[0]

        if not canceled_order.result.value == "Ok":
            raise ValueError(
                f"Failed to cancel order {order_id}: {canceled_order.result.value}"
            )

        new_order = await self.create_order(
            order_type=old_order.type.value,
            quantity=quantity,
            side=old_order.side.value,
            price=price,
            product_id=old_order.product_id,
            sender=old_order.sender,
            subaccount=subaccount.name,
            time_in_force=time_in_force or self.default_time_in_force,
            post_only=post_only or self.default_post_only,
            reduce_only=reduce_only,
            dry_run=False,
        )
        return self._models.SubmitOrderCreatedDto.model_validate(
            new_order
        ), canceled_order.result.value == "Ok"



import time
import uuid
from eth_utils import add_0x_prefix, decode_hex, encode_hex


class Bytes32Error(ValueError):
    pass


def ensure_bytes32(value):
    """Return a 32-byte representation of ``value`` (zero-padded on the left)."""
    if isinstance(value, (bytes, bytearray)):
        data = bytes(value)
    elif isinstance(value, str):
        raw = value[2:] if value.startswith("0x") else value
        raw = raw if len(raw) % 2 == 0 else "0" + raw
        try:
            data = decode_hex(add_0x_prefix(raw))
        except ValueError as exc:
            raise Bytes32Error(f"Invalid hex string: {value!r}") from exc
    else:
        raise Bytes32Error(f"Unsupported type: {type(value).__name__}")
    if len(data) > 32:
        raise Bytes32Error(f"Expected at most 32 bytes, got {len(data)} bytes")
    return data.rjust(32, b"\x00")


def ensure_bytes32_hex(value):
    """Return ``value`` as a 0x-prefixed 64-hex-digit string."""
    return encode_hex(ensure_bytes32(value))


def uuid_to_bytes32(uuid_str: str) -> str:
    """Converts UUID string to bytes32 hex format.

    Args:
        uuid_str (str): UUID string to convert.

    Returns:
        str: Bytes32 hex string prefixed with '0x'.
    """
    uuid_obj = uuid.UUID(uuid_str)

    # remove hyphens and convert to hex
    uuid_hex = uuid_obj.hex

    # pad the hex to make it 32 bytes
    padded_hex = uuid_hex.rjust(64, "0")

    return "0x" + padded_hex


def is_uuid(value: str) -> bool:
    """Checks if a string is a valid UUID.

    Args:
        value (str): String to check.

    Returns:
        bool: True if string is a valid UUID, False otherwise.
    """
    try:
        return value == str(uuid.UUID(value))
    except ValueError:
        return False


def client_order_id_to_bytes32(client_order_id: str) -> str:
    """Converts client_order_id to appropriate bytes32 format.

    Args:
        client_order_id (str): Client order ID to convert.

    Returns:
        str: Converted client order ID in bytes32 hex format.

    Raises:
        ValueError: If string is longer than 32 characters and not a UUID, or if input is None/empty.
    """
    if client_order_id is None:
        raise ValueError("Client order ID cannot be None")

    if not client_order_id:
        raise ValueError("Client order ID cannot be empty")

    if is_uuid(client_order_id):
        return uuid_to_bytes32(client_order_id)

    if len(client_order_id) > 32:
        raise ValueError(
            f"Client order ID cannot be longer than 32 characters, got {len(client_order_id)}"
        )

    # Convert string to bytes32 hex format
    client_order_bytes = client_order_id.encode("utf-8")
    padded_bytes = client_order_bytes.ljust(32, b"\0")
    return "0x" + padded_bytes.hex()


def generate_nonce() -> str:
    """Generates a timestamp-based nonce.

    Returns:
        str: Current timestamp in nanoseconds as string.
    """
    return str(time.time_ns())



import os
import pytest
from dotenv import load_dotenv
from eth_account import Account
import ethereal
from ethereal.async_rest_client import AsyncRESTClient
import pytest_asyncio
from ethereal.ws.ws_base import WSBase

# Force reload of the dotenv file.
load_dotenv(override=True)

# Network configurations with hardcoded URLs
NETWORKS = {
    "devnet": {
        "chain_id": 13374201,
        "base_url": "https://api.etherealdev.net/",
        "rpc_url": "https://rpc.etherealdev.net",
        "ws_url": "wss://ws.etherealdev.net/",
    },
    "testnet": {
        "chain_id": 13374202,
        "base_url": "https://api.etherealtest.net/",
        "rpc_url": "https://rpc.etherealtest.net",
        "ws_url": "wss://ws.etherealtest.net/",
    },
    "mainnet": {
        "chain_id": 5064014,
        "base_url": "https://api.ethereal.trade/",
        "rpc_url": "https://rpc.ethereal.trade",
        "ws_url": "wss://ws.ethereal.trade/",
    },
}


def pytest_addoption(parser):
    """Add --network option to pytest."""
    parser.addoption(
        "--network",
        action="store",
        default=os.getenv("NETWORK", "testnet"),
        choices=list(NETWORKS.keys()),
        help="Network to test against (default: testnet)",
    )


def pytest_collection_modifyitems(config, items):
    """Skip tests based on network markers."""
    network = config.getoption("--network")

    for item in items:
        # Skip tests marked for specific networks only
        if item.get_closest_marker(f"{network}_only") is None:
            # Check if it's marked for another network
            for net in NETWORKS:
                if net != network and item.get_closest_marker(f"{net}_only"):
                    item.add_marker(pytest.mark.skip(reason=f"Only runs on {net}"))

        # Skip tests marked to skip on current network
        if item.get_closest_marker(f"skip_{network}"):
            item.add_marker(pytest.mark.skip(reason=f"Skipped on {network}"))


@pytest.fixture(scope="session")
def network(request):
    """Current network being tested."""
    return request.config.getoption("--network")


@pytest.fixture(scope="session")
def network_config(network):
    """Network configuration for current network."""
    return NETWORKS[network]


@pytest.fixture(scope="session")
def private_key():
    """Get private key from environment."""
    key = os.getenv("PRIVATE_KEY")
    if not key:
        pytest.fail("PRIVATE_KEY environment variable is required")
    return key


@pytest.fixture(scope="session")
def rc(network, network_config, private_key):
    """REST client for testing."""
    config = {
        "network": network,
        "base_url": network_config["base_url"],
        "chain_config": {
            "chain_id": network_config["chain_id"],
            "rpc_url": network_config["rpc_url"],
            "private_key": private_key,
        },
    }
    rc = ethereal.RESTClient(config)
    assert rc is not None
    assert rc.chain is not None
    try:
        yield rc
    finally:
        rc.close()


@pytest.fixture(scope="session")
def rc_ro(network, network_config, private_key):
    """Read-only REST client for testing."""
    account = Account.from_key(private_key)
    address = account.address

    config = {
        "network": network,
        "base_url": network_config["base_url"],
        "chain_config": {
            "chain_id": network_config["chain_id"],
            "rpc_url": network_config["rpc_url"],
            "address": address,
        },
    }
    rc = ethereal.RESTClient(config)
    assert rc is not None
    assert rc.chain is not None
    return rc


@pytest.fixture(scope="session")
def ws_base(network_config):
    """Base WebSocket client for testing."""
    config = {
        "base_url": network_config["ws_url"],
    }
    ws = WSBase(config)
    ws.open(namespaces=["/", "/v1/stream"])
    assert ws is not None
    try:
        yield ws
    finally:
        ws.close()


@pytest.fixture(scope="session")
def ws(network_config):
    """WebSocket client for testing."""
    config = {
        "base_url": network_config["ws_url"],
    }
    ws = ethereal.WSClient(config)
    ws.open(namespaces=["/", "/v1/stream"])
    assert ws is not None
    try:
        yield ws
    finally:
        ws.close()


@pytest.fixture(scope="session")
def sid(rc):
    """Return the subaccount id for the current test."""
    if len(rc.subaccounts) == 0:
        raise ValueError("No subaccounts found for the connected address.")
    return rc.subaccounts[0].id


@pytest.fixture(scope="session")
def sname(rc):
    """Return the subaccount name for the current test."""
    if len(rc.subaccounts) == 0:
        raise ValueError("No subaccounts found for the connected address.")
    return rc.subaccounts[0].name


@pytest_asyncio.fixture
async def async_rc(network, network_config, private_key):
    """Asynchronous REST client for testing."""
    config = {
        "network": network,
        "base_url": network_config["base_url"],
        "chain_config": {
            "chain_id": network_config["chain_id"],
            "rpc_url": network_config["rpc_url"],
            "private_key": private_key,
        },
    }
    rc = await AsyncRESTClient.create(config)
    assert rc is not None
    assert rc.chain is not None
    try:
        yield rc
    finally:
        await rc.close()


@pytest_asyncio.fixture
async def async_rc_ro(network, network_config, private_key):
    """Read-only asynchronous REST client for testing."""
    account = Account.from_key(private_key)
    address = account.address

    config = {
        "network": network,
        "base_url": network_config["base_url"],
        "chain_config": {
            "chain_id": network_config["chain_id"],
            "rpc_url": network_config["rpc_url"],
            "address": address,
        },
    }
    rc = await AsyncRESTClient.create(config)
    assert rc is not None
    assert rc.chain is not None
    try:
        yield rc
    finally:
        await rc.close()


@pytest_asyncio.fixture
async def async_subaccount(async_rc):
    """Get first subaccount for async client."""
    subaccounts = await async_rc.list_subaccounts(sender=async_rc.chain.address)
    if not subaccounts:
        raise ValueError("No subaccounts found for the connected address.")
    return subaccounts[0]


@pytest_asyncio.fixture
async def async_subaccount_ro(async_rc_ro):
    """Get first subaccount for read-only async client."""
    subaccounts = await async_rc_ro.list_subaccounts(sender=async_rc_ro.chain.address)
    if not subaccounts:
        raise ValueError("No subaccounts found for the connected address.")
    return subaccounts[0]



from ethereal.__version__ import __version__

USER_AGENT = f"ethereal-py-sdk/{__version__}"
PRIVATE_KEY = "PRIVATE_KEY"
RPC_URL = "RPC_URL"

BASE_URL = "https://api.etherealtest.net"
API_PREFIX = "/v1"

WS_BASE_URL = "wss://ws.etherealtest.net"

# Network URLs
NETWORK_URLS = {
    "mainnet": "https://api.ethereal.trade",
    "testnet": "https://api.etherealtest.net",
    "devnet": "https://api.etherealdev.net",
}

ARCHIVE_NETWORK_URLS = {
    "mainnet": "https://archive.ethereal.trade",
    "testnet": "https://archive.etherealtest.net",
    "devnet": "https://archive.etherealdev.net",
}
WS_NAMESPACES = [
    "/",
    "/v1/stream",
]

X_RATELIMIT_LIMIT = "x-ratelimit-limit"
X_RATELIMIT_REMAINING = "x-ratelimit-remaining"
RETRY_AFTER = "retry-after"
RATE_LIMIT_HEADERS = {X_RATELIMIT_LIMIT, X_RATELIMIT_REMAINING, RETRY_AFTER}
REST_COMMON_FIELDS = {
    X_RATELIMIT_LIMIT: "rate_limit_limit",
    X_RATELIMIT_REMAINING: "rate_limit_remaining",
    RETRY_AFTER: "retry_after",
}



from typing import List

from ethereal.constants import API_PREFIX
from ethereal.models.rest import ProductDto, MarketLiquidityDto, MarketPriceDto


async def list_products(self, **kwargs) -> List[ProductDto]:
    """Lists all products and their configurations.

    Other Parameters:
        order (str, optional): Sort order, 'asc' or 'desc'. Optional.
        limit (float, optional): Maximum number of results to return. Optional.
        cursor (str, optional): Pagination cursor for fetching the next page. Optional.
        order_by (str, optional): Field to order by (e.g., 'createdAt'). Optional.
        ticker (str, optional): Filter by product ticker (e.g., 'ETHUSD'). Optional.
        **kwargs: Additional request parameters accepted by the API. Optional.

    Returns:
        List[ProductDto]: Product configuration records.
    """
    res = await self.get_validated(
        url_path=f"{API_PREFIX}/product",
        request_model=self._models.V1ProductGetParametersQuery,
        response_model=self._models.PageOfProductDtos,
        **kwargs,
    )
    data = [
        self._models.ProductDto(**model.model_dump(by_alias=True)) for model in res.data
    ]
    return data


async def get_market_liquidity(self, **kwargs) -> MarketLiquidityDto:
    """Gets market liquidity for a product.

    Args:
        product_id (str): Id representing the registered product. Required.

    Other Parameters:
        **kwargs: Additional request parameters accepted by the API. Optional.

    Returns:
        MarketLiquidityDto: Top-of-book and depth liquidity metrics.
    """
    res = await self.get_validated(
        url_path=f"{API_PREFIX}/product/market-liquidity",
        request_model=self._models.V1ProductMarketLiquidityGetParametersQuery,
        response_model=self._models.MarketLiquidityDto,
        **kwargs,
    )
    return res


async def list_market_prices(self, **kwargs) -> List[MarketPriceDto]:
    """Gets market prices for one or more products.

    Args:
        product_ids (List[str]): List of product IDs to query. Required.

    Other Parameters:
        **kwargs: Additional request parameters accepted by the API. Optional.

    Returns:
        List[MarketPriceDto]: Best bid/ask prices for the requested products.
    """
    res = await self.get_validated(
        url_path=f"{API_PREFIX}/product/market-price",
        request_model=self._models.V1ProductMarketPriceGetParametersQuery,
        response_model=self._models.ListOfMarketPriceDtos,
        **kwargs,
    )
    data = [
        self._models.MarketPriceDto(**model.model_dump(by_alias=True))
        for model in res.data
    ]
    return data



import pytest
import asyncio


def test_ws_open(ws_base):
    """Test opening a websocket connection."""
    assert ws_base.sio is not None


@pytest.mark.asyncio
async def test_ws_emit(ws_base):
    """Test emitting an event to the server."""

    def log_callback(data):
        ws_base.logger.info(f"Received data: {data}")

    ws_base.callbacks["BookDepth"] = [log_callback]
    ws_base._emit(
        "subscribe",
        {"type": "BookDepth", "productId": "4368f737-eeed-40bc-bb1b-7b9bfd88a057"},
        namespace="/v1/stream",
    )

    # await until a log appears
    await asyncio.sleep(2)
    assert ws_base.sio is not None



from __future__ import annotations

import asyncio
import inspect
import threading
from concurrent.futures import Future
from typing import Any, Callable, Dict, Optional, Union

from ethereal.async_rest_client import AsyncRESTClient
from ethereal.models.config import RESTConfig

try:
    import uvloop  # type: ignore

    uvloop_enabled = True
except Exception:
    uvloop_enabled = False


def _create_event_loop() -> asyncio.AbstractEventLoop:
    if uvloop_enabled:
        try:
            asyncio.set_event_loop_policy(uvloop.EventLoopPolicy())
        except Exception:
            pass
    return asyncio.new_event_loop()


class RESTClient:
    """Synchronous wrapper around AsyncRESTClient.

    Reuses a single async client on a background event loop thread and keeps
    backward compatibility (e.g., exposes `_async_client`). Use
    `AsyncRESTClient` directly in async code.
    """

    def __init__(self, config: Union[Dict[str, Any], RESTConfig] = {}):
        try:
            asyncio.get_running_loop()
            raise RuntimeError(
                "RESTClient cannot be used in async context. Use AsyncRESTClient instead."
            )
        except RuntimeError as e:
            if "RESTClient cannot be used in async context" in str(e):
                raise

        self._config = config
        self._loop: Optional[asyncio.AbstractEventLoop] = None
        self._thread: Optional[threading.Thread] = None
        self._closed = False
        self._start_loop_thread()
        self._async_client: AsyncRESTClient = self._run_sync(
            lambda: AsyncRESTClient.create(self._config)
        )

    def _start_loop_thread(self):
        if self._thread is not None:
            return

        loop_ready = threading.Event()

        def _run_loop():
            self._loop = _create_event_loop()
            asyncio.set_event_loop(self._loop)
            loop_ready.set()
            self._loop.run_forever()

        self._thread = threading.Thread(
            target=_run_loop, name="EtherealRESTClientLoop", daemon=True
        )
        self._thread.start()
        loop_ready.wait()

    def _submit(self, coro_fn: Callable[[], Any]) -> Future:
        if self._closed:
            raise RuntimeError("RESTClient is closed")
        if self._loop is None:
            raise RuntimeError("Event loop not initialized")
        result = coro_fn()
        if not inspect.iscoroutine(result):

            async def _wrap(v):
                return v

            result = _wrap(result)
        return asyncio.run_coroutine_threadsafe(result, self._loop)

    def _run_sync(self, coro_factory: Callable[[], Any]):
        fut = self._submit(coro_factory)
        return fut.result()

    def __getattr__(self, name: str):
        attr = getattr(self._async_client, name, None)
        if attr is None:
            raise AttributeError(name)

        auto_call_methods = {
            "subaccounts",
            "products",
            "tokens",
            "products_by_id",
            "products_by_ticker",
        }

        if name in auto_call_methods and inspect.iscoroutinefunction(attr):
            return self._run_sync(lambda: attr())

        if inspect.iscoroutinefunction(attr):

            def _sync_wrapper(*args, **kwargs):
                return self._run_sync(lambda: attr(*args, **kwargs))

            return _sync_wrapper

        if callable(attr):

            def _callable_wrapper(*args, **kwargs):
                return self._run_sync(lambda: attr(*args, **kwargs))

            return _callable_wrapper

        return self._run_sync(lambda: getattr(self._async_client, name))

    def close(self):
        if self._closed:
            return
        self._closed = True

        try:
            if self._loop is not None:
                try:
                    self._run_sync(lambda: self._async_client.close())
                except Exception:
                    pass

                self._loop.call_soon_threadsafe(self._loop.stop)
        finally:
            if self._thread is not None:
                self._thread.join(timeout=5)
            self._thread = None
            self._loop = None

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close()



"""Pure async tests for OTO/OCO order patterns."""

import uuid
import asyncio
import pytest
import time


def safe_round_price(price, tick_size):
    """Round price to valid tick size using decimal precision."""
    from decimal import Decimal as D, ROUND_HALF_UP

    if tick_size == 0:
        return float(price)

    price_decimal = D(str(price))
    tick_size_decimal = D(str(tick_size))
    return float(
        (price_decimal / tick_size_decimal).quantize(D("1"), rounding=ROUND_HALF_UP)
        * tick_size_decimal
    )


@pytest.mark.asyncio
async def test_rest_pure_oto_pattern(async_rc, network):
    subaccounts = await async_rc.list_subaccounts(sender=async_rc.chain.address)
    sub = subaccounts[0]
    products = await async_rc.list_products()
    pid = products[0].id
    tick_size = float((await async_rc.products_by_id())[pid].tick_size)
    prices = (await async_rc.list_market_prices(product_ids=[pid]))[0]
    best_bid = float(prices.best_bid_price)
    group_id = uuid.uuid4()

    entry_price = safe_round_price(best_bid * 0.95, tick_size)
    primary = await async_rc.create_order(
        order_type="LIMIT",
        product_id=pid,
        side=0,
        price=entry_price,
        quantity=0.001,
        time_in_force="GTD",
        group_id=group_id,
        group_contingency_type=0,
        sender=async_rc.chain.address,
        subaccount=sub.name,
    )
    assert isinstance(primary, async_rc._models.SubmitOrderCreatedDto)

    latest = (await async_rc.list_market_prices(product_ids=[pid]))[0]
    exit_price = safe_round_price(float(latest.best_bid_price) * 1.05, tick_size)
    secondary = await async_rc.create_order(
        order_type="LIMIT",
        product_id=pid,
        side=1,
        price=exit_price,
        quantity=0.001,
        time_in_force="GTD",
        group_id=group_id,
        sender=async_rc.chain.address,
        subaccount=sub.name,
    )
    assert isinstance(secondary, async_rc._models.SubmitOrderCreatedDto)

    await asyncio.sleep(2)
    p = await async_rc.get_order(id=primary.id)
    s = await async_rc.get_order(id=secondary.id)
    assert p.group_id == group_id and p.group_contingency_type.value == 0
    assert s.group_id == group_id

    await async_rc.cancel_orders(
        sender=async_rc.chain.address,
        subaccount=sub.name,
        order_ids=[primary.id, secondary.id],
    )


@pytest.mark.asyncio
async def test_rest_pure_oco_pattern(async_rc, network):
    subaccounts = await async_rc.list_subaccounts(sender=async_rc.chain.address)
    sub = subaccounts[0]
    products = await async_rc.list_products()
    pid = products[0].id
    tick = float((await async_rc.products_by_id())[pid].tick_size)
    prices = (await async_rc.list_market_prices(product_ids=[pid]))[0]
    best_bid = float(prices.best_bid_price)
    best_ask = float(prices.best_ask_price)
    group_id = uuid.uuid4()

    market = await async_rc.create_order(
        order_type="MARKET",
        product_id=pid,
        side=0,
        quantity=0.001,
        time_in_force="GTD",
        sender=async_rc.chain.address,
        subaccount=sub.name,
    )
    assert isinstance(market, async_rc._models.SubmitOrderCreatedDto)

    sl_price = safe_round_price(best_bid * 0.90, tick)
    sl = await async_rc.create_order(
        order_type="LIMIT",
        product_id=pid,
        side=1,
        price=sl_price,
        quantity=0.001,
        time_in_force="GTD",
        reduce_only=True,
        group_id=group_id,
        group_contingency_type=1,
        stop_type=1,
        stop_price=sl_price,
        sender=async_rc.chain.address,
        subaccount=sub.name,
    )
    assert isinstance(sl, async_rc._models.SubmitOrderCreatedDto)

    tp_price = safe_round_price(best_ask * 1.10, tick)
    tp = await async_rc.create_order(
        order_type="LIMIT",
        product_id=pid,
        side=1,
        price=tp_price,
        quantity=0.001,
        time_in_force="GTD",
        reduce_only=True,
        group_id=group_id,
        group_contingency_type=1,
        stop_type=0,
        stop_price=tp_price,
        sender=async_rc.chain.address,
        subaccount=sub.name,
    )
    assert isinstance(tp, async_rc._models.SubmitOrderCreatedDto)

    await asyncio.sleep(2)
    sl_o = await async_rc.get_order(id=sl.id)
    tp_o = await async_rc.get_order(id=tp.id)
    assert sl_o.group_id == group_id and sl_o.group_contingency_type.value == 1
    assert tp_o.group_id == group_id and tp_o.group_contingency_type.value == 1
    assert sl_o.status.value in [
        "NEW",
        "PENDING",
        "CANCELED",
    ] and tp_o.status.value in ["NEW", "PENDING", "CANCELED"]

    await async_rc.cancel_orders(
        sender=async_rc.chain.address, subaccount=sub.name, order_ids=[sl.id, tp.id]
    )
    close = await async_rc.create_order(
        order_type="MARKET",
        product_id=pid,
        quantity=0,
        side=1,
        time_in_force="GTD",
        reduce_only=True,
        close=True,
        sender=async_rc.chain.address,
        subaccount=sub.name,
    )
    assert isinstance(close, async_rc._models.SubmitOrderCreatedDto)


@pytest.mark.asyncio
async def test_rest_otoco_pattern_complete(async_rc, network):
    subaccounts = await async_rc.list_subaccounts(sender=async_rc.chain.address)
    sub = subaccounts[0]
    products = await async_rc.list_products()
    pid = products[0].id
    tick = float((await async_rc.products_by_id())[pid].tick_size)
    prices = (await async_rc.list_market_prices(product_ids=[pid]))[0]
    best_bid = float(prices.best_bid_price)
    best_ask = float(prices.best_ask_price)
    group_id = uuid.uuid4()

    entry_price = safe_round_price(best_bid * 0.95, tick)
    entry = await async_rc.create_order(
        order_type="LIMIT",
        product_id=pid,
        side=0,
        price=entry_price,
        quantity=0.002,
        time_in_force="GTD",
        group_id=group_id,
        group_contingency_type=0,
        sender=async_rc.chain.address,
        subaccount=sub.name,
    )
    assert isinstance(entry, async_rc._models.SubmitOrderCreatedDto)

    tp_price = safe_round_price(best_ask * 1.10, tick)
    tp = await async_rc.create_order(
        order_type="LIMIT",
        product_id=pid,
        side=1,
        price=tp_price,
        quantity=0.002,
        time_in_force="GTD",
        reduce_only=True,
        group_id=group_id,
        sender=async_rc.chain.address,
        subaccount=sub.name,
    )
    assert isinstance(tp, async_rc._models.SubmitOrderCreatedDto)

    sl_price = safe_round_price(best_bid * 0.90, tick)
    sl = await async_rc.create_order(
        order_type="LIMIT",
        product_id=pid,
        side=1,
        price=sl_price,
        quantity=0.002,
        time_in_force="GTD",
        reduce_only=True,
        group_id=group_id,
        sender=async_rc.chain.address,
        subaccount=sub.name,
    )
    assert isinstance(sl, async_rc._models.SubmitOrderCreatedDto)

    await asyncio.sleep(2)
    e = await async_rc.get_order(id=entry.id)
    assert e.group_id == group_id and e.group_contingency_type.value == 0
    await async_rc.cancel_orders(
        sender=async_rc.chain.address,
        subaccount=sub.name,
        order_ids=[entry.id, tp.id, sl.id],
    )


@pytest.mark.asyncio
async def test_rest_stop_orders_with_oco(async_rc, network):
    subaccounts = await async_rc.list_subaccounts(sender=async_rc.chain.address)
    sub = subaccounts[0]
    products = await async_rc.list_products()
    pid = products[0].id
    tick = float((await async_rc.products_by_id())[pid].tick_size)
    prices = (await async_rc.list_market_prices(product_ids=[pid]))[0]
    best_bid = float(prices.best_bid_price)
    best_ask = float(prices.best_ask_price)
    group_id = uuid.uuid4()

    # Place stop-loss and take-profit as OCO
    sl_trigger = safe_round_price(best_bid * 0.90, tick)
    sl = await async_rc.create_order(
        order_type="LIMIT",
        product_id=pid,
        side=1,
        price=sl_trigger,
        quantity=0.001,
        time_in_force="GTD",
        reduce_only=True,
        group_id=group_id,
        group_contingency_type=1,
        stop_type=1,
        stop_price=sl_trigger,
        sender=async_rc.chain.address,
        subaccount=sub.name,
    )
    assert isinstance(sl, async_rc._models.SubmitOrderCreatedDto)

    tp_trigger = safe_round_price(best_ask * 1.10, tick)
    tp = await async_rc.create_order(
        order_type="LIMIT",
        product_id=pid,
        side=1,
        price=tp_trigger,
        quantity=0.001,
        time_in_force="GTD",
        reduce_only=True,
        group_id=group_id,
        group_contingency_type=1,
        stop_type=0,
        stop_price=tp_trigger,
        sender=async_rc.chain.address,
        subaccount=sub.name,
    )
    assert isinstance(tp, async_rc._models.SubmitOrderCreatedDto)

    await asyncio.sleep(2)
    tp_o = await async_rc.get_order(id=tp.id)
    sl_o = await async_rc.get_order(id=sl.id)
    assert (
        tp_o.reduce_only
        and tp_o.group_id == group_id
        and tp_o.group_contingency_type.value == 1
    )
    assert (
        sl_o.reduce_only
        and sl_o.group_id == group_id
        and sl_o.group_contingency_type.value == 1
    )
    await async_rc.cancel_orders(
        sender=async_rc.chain.address, subaccount=sub.name, order_ids=[tp.id, sl.id]
    )


@pytest.mark.asyncio
async def test_rest_time_in_force_with_oto_oco(async_rc, network):
    subaccounts = await async_rc.list_subaccounts(sender=async_rc.chain.address)
    sub = subaccounts[0]
    products = await async_rc.list_products()
    pid = products[0].id
    tick = float((await async_rc.products_by_id())[pid].tick_size)
    prices = (await async_rc.list_market_prices(product_ids=[pid]))[0]
    best_bid = float(prices.best_bid_price)
    best_ask = float(prices.best_ask_price)
    group_id = uuid.uuid4()

    expires_at = int(time.time()) + 3600

    entry_price = safe_round_price(best_bid * 0.95, tick)
    entry = await async_rc.create_order(
        order_type="LIMIT",
        product_id=pid,
        side=0,
        price=entry_price,
        quantity=0.001,
        time_in_force="GTD",
        expires_at=expires_at,
        group_id=group_id,
        group_contingency_type=0,
        sender=async_rc.chain.address,
        subaccount=sub.name,
    )
    assert isinstance(entry, async_rc._models.SubmitOrderCreatedDto)

    exit_price = safe_round_price(best_ask * 1.05, tick)
    exit_o = await async_rc.create_order(
        order_type="LIMIT",
        product_id=pid,
        side=1,
        price=exit_price,
        quantity=0.001,
        time_in_force="GTD",
        group_id=group_id,
        sender=async_rc.chain.address,
        subaccount=sub.name,
    )
    assert isinstance(exit_o, async_rc._models.SubmitOrderCreatedDto)

    await asyncio.sleep(2)
    e = await async_rc.get_order(id=entry.id)
    x = await async_rc.get_order(id=exit_o.id)
    assert e.time_in_force.value == "GTD" and e.expires_at is not None
    assert x.time_in_force.value == "GTD"
    await async_rc.cancel_orders(
        sender=async_rc.chain.address,
        subaccount=sub.name,
        order_ids=[entry.id, exit_o.id],
    )


@pytest.mark.asyncio
async def test_rest_post_only_with_oco(async_rc, network):
    subaccounts = await async_rc.list_subaccounts(sender=async_rc.chain.address)
    sub = subaccounts[0]
    products = await async_rc.list_products()
    pid = products[0].id
    tick = float((await async_rc.products_by_id())[pid].tick_size)
    prices = (await async_rc.list_market_prices(product_ids=[pid]))[0]
    best_bid = float(prices.best_bid_price)
    best_ask = float(prices.best_ask_price)
    group_id = uuid.uuid4()

    buy_price = safe_round_price(best_bid * 0.99, tick)
    buy = await async_rc.create_order(
        order_type="LIMIT",
        product_id=pid,
        side=0,
        price=buy_price,
        quantity=0.001,
        time_in_force="GTD",
        post_only=True,
        group_id=group_id,
        group_contingency_type=1,
        sender=async_rc.chain.address,
        subaccount=sub.name,
    )
    assert isinstance(buy, async_rc._models.SubmitOrderCreatedDto)

    sell_price = safe_round_price(best_ask * 1.01, tick)
    sell = await async_rc.create_order(
        order_type="LIMIT",
        product_id=pid,
        side=1,
        price=sell_price,
        quantity=0.001,
        time_in_force="GTD",
        post_only=True,
        group_id=group_id,
        group_contingency_type=1,
        sender=async_rc.chain.address,
        subaccount=sub.name,
    )
    assert isinstance(sell, async_rc._models.SubmitOrderCreatedDto)

    await asyncio.sleep(2)
    b = await async_rc.get_order(id=buy.id)
    s = await async_rc.get_order(id=sell.id)
    assert b.post_only and b.group_id == group_id
    assert s.post_only and s.group_id == group_id
    await async_rc.cancel_orders(
        sender=async_rc.chain.address, subaccount=sub.name, order_ids=[buy.id, sell.id]
    )



"""Pure sync tests for REST info endpoints."""

from typing import List
from pydantic import BaseModel


def test_rest_subaccounts(rc):
    subs = rc.list_subaccounts(sender=rc.chain.address)
    assert isinstance(subs, List)
    assert all(isinstance(sa, BaseModel) for sa in subs)


def test_rest_subaccount(rc):
    subs = rc.list_subaccounts(sender=rc.chain.address)
    sub = rc.get_subaccount(subs[0].id)
    assert isinstance(sub, BaseModel)


def test_rest_rpc_config(rc):
    cfg = rc.get_rpc_config()
    assert isinstance(cfg, BaseModel)


def test_rest_products(rc):
    products = rc.list_products()
    assert isinstance(products, List)
    assert all(isinstance(p, BaseModel) for p in products)


def test_rest_tokens(rc):
    tokens = rc.list_tokens()
    assert isinstance(tokens, List)
    assert all(isinstance(t, BaseModel) for t in tokens)


def test_rest_market_prices(rc):
    products = rc.list_products()
    if products:
        prices = rc.list_market_prices(product_ids=[products[0].id])
        assert isinstance(prices, List)
        assert all(isinstance(p, BaseModel) for p in prices)



from typing import List
from uuid import UUID
from ethereal.constants import API_PREFIX
from ethereal.models.rest import SubaccountDto, SubaccountBalanceDto


async def list_subaccounts(self, **kwargs) -> List[SubaccountDto]:
    """Lists subaccounts for a given sender (address).

    Args:
        sender (str): Wallet address to query subaccounts for. Required.

    Other Parameters:
        name (str, optional): Filter by subaccount name. Optional.
        order (str, optional): Sort order, 'asc' or 'desc'. Optional.
        limit (float, optional): Maximum number of results to return. Optional.
        cursor (str, optional): Pagination cursor for fetching the next page. Optional.
        order_by (str, optional): Field to order by (e.g., 'createdAt'). Optional.
        **kwargs: Additional request parameters accepted by the API. Optional.

    Returns:
        List[SubaccountDto]: Subaccount records for the sender.
    """
    res = await self.get_validated(
        url_path=f"{API_PREFIX}/subaccount/",
        request_model=self._models.V1SubaccountGetParametersQuery,
        response_model=self._models.PageOfSubaccountDtos,
        **kwargs,
    )
    data = [
        self._models.SubaccountDto(**model.model_dump(by_alias=True))
        for model in res.data
    ]
    return data


async def get_subaccount(self, id: UUID, **kwargs) -> SubaccountDto:
    """Gets a specific subaccount by ID.

    Args:
        id (str): UUID of the subaccount. Required.

    Other Parameters:
        **kwargs: Additional request parameters accepted by the API. Optional.

    Returns:
        SubaccountDto: Subaccount details.
    """
    endpoint = f"{API_PREFIX}/subaccount/{id}"
    res = await self.get(endpoint, **kwargs)
    return self._models.SubaccountDto(**res)


async def get_subaccount_balances(self, **kwargs) -> List[SubaccountBalanceDto]:
    """Gets token balances for a subaccount.

    Args:
        subaccount_id (str): UUID of the subaccount. Required.

    Other Parameters:
        order (str, optional): Sort order, 'asc' or 'desc'. Optional.
        limit (float, optional): Maximum number of results to return. Optional.
        cursor (str, optional): Pagination cursor for fetching the next page. Optional.
        order_by (str, optional): Field to order by (e.g., 'tokenName', 'updatedAt', 'amount'). Optional.
        **kwargs: Additional request parameters accepted by the API. Optional.

    Returns:
        List[SubaccountBalanceDto]: Balances for the subaccount.
    """
    res = await self.get_validated(
        url_path=f"{API_PREFIX}/subaccount/balance",
        request_model=self._models.V1SubaccountBalanceGetParametersQuery,
        response_model=self._models.PageOfSubaccountBalanceDtos,
        **kwargs,
    )
    data = [
        self._models.SubaccountBalanceDto(**model.model_dump(by_alias=True))
        for model in res.data
    ]
    return data


async def get_subaccount_balance_history(self, **kwargs) -> List["BalanceHistoryDto"]:
    """Gets historical subaccount balances from the archive API.

    Args:
        subaccount_id (str): UUID of the subaccount. Required.

    Other Parameters:
        start_time (float): Range start time in milliseconds since Unix epoch. Required.
        end_time (float, optional): Range end time in milliseconds since Unix epoch. Optional.
        resolution (str): Data resolution (e.g., 'hour1', 'day1'). Required.
        order (str, optional): Sort order, 'asc' or 'desc'. Optional.
        limit (float, optional): Maximum number of results to return. Optional.
        cursor (str, optional): Pagination cursor for fetching the next page. Optional.
        order_by (str, optional): Field to order by (defaults to 'time'). Optional.
        **kwargs: Additional query parameters supported by the API.

    Returns:
        List[BalanceHistoryDto]: Historical balance records ordered per request parameters.
    """

    res = await self.get_validated(
        url_path=f"{API_PREFIX}/subaccount/balance",
        request_model=self._models.ArchiveV1SubaccountBalanceGetParametersQuery,
        response_model=self._models.PageOfBalanceHistoryDtos,
        base_url_override=self._archive_base_url,
        **kwargs,
    )

    data = [
        self._models.BalanceHistoryDto(**model.model_dump(by_alias=True))
        for model in res.data
    ]
    return data


async def get_subaccount_unrealized_pnl_history(
    self, **kwargs
) -> List["UnrealizedPnlHistoryDto"]:
    """Gets historical unrealized PnL for a subaccount from the archive API."""

    res = await self.get_validated(
        url_path=f"{API_PREFIX}/subaccount/unrealized-pnl",
        request_model=self._models.ArchiveV1SubaccountUnrealizedPnlGetParametersQuery,
        response_model=self._models.PageOfUnrealizedPnlHistoryDtos,
        base_url_override=self._archive_base_url,
        **kwargs,
    )

    data = [
        self._models.UnrealizedPnlHistoryDto(**model.model_dump(by_alias=True))
        for model in res.data
    ]
    return data


async def get_subaccount_volume_history(self, **kwargs) -> List["VolumeHistoryDto"]:
    """Gets historical trading volume for a subaccount from the archive API."""

    res = await self.get_validated(
        url_path=f"{API_PREFIX}/subaccount/volume",
        request_model=self._models.ArchiveV1SubaccountVolumeGetParametersQuery,
        response_model=self._models.PageOfVolumeHistoryDtos,
        base_url_override=self._archive_base_url,
        **kwargs,
    )

    data = [
        self._models.VolumeHistoryDto(**model.model_dump(by_alias=True))
        for model in res.data
    ]
    return data


async def get_subaccount_funding_history(
    self, **kwargs
) -> List["PositionFundingHistoryDto"]:
    """Gets historical funding charges for positions in a subaccount from the archive API.

    Args:
        subaccount_id (str): UUID of the subaccount. Required.
        start_time (float): Range start time in milliseconds since Unix epoch. Required.

    Other Parameters:
        end_time (float, optional): Range end time in milliseconds since Unix epoch. Optional.
        position_ids (List[str], optional): Filter by specific position IDs (max 128). Optional.
        product_ids (List[str], optional): Filter by specific product IDs. Optional.
        order (str, optional): Sort order, 'asc' or 'desc'. Optional.
        limit (float, optional): Maximum number of results to return. Optional.
        cursor (str, optional): Pagination cursor for fetching the next page. Optional.
        order_by (str, optional): Field to order by (defaults to 'time'). Optional.
        **kwargs: Additional query parameters supported by the API.

    Returns:
        List[PositionFundingHistoryDto]: Historical funding charge records for positions in the subaccount.
    """

    res = await self.get_validated(
        url_path=f"{API_PREFIX}/subaccount/funding",
        request_model=self._models.ArchiveV1SubaccountFundingGetParametersQuery,
        response_model=self._models.PageOfPositionFundingHistoryDtos,
        base_url_override=self._archive_base_url,
        **kwargs,
    )

    data = [
        self._models.PositionFundingHistoryDto(**model.model_dump(by_alias=True))
        for model in res.data
    ]
    return data



from typing import Any, Dict, Optional, Union, TypedDict, Type
from pydantic import BaseModel, ValidationError

import requests
from requests.exceptions import HTTPError

from ethereal.base_client import BaseClient
from ethereal.models.config import HTTPConfig
from ethereal.constants import RATE_LIMIT_HEADERS, USER_AGENT, REST_COMMON_FIELDS


class ValidationException(Exception):
    def __init__(self, stage: str, url_path: str, model: Type[BaseModel], errors: list):
        def alias_to_name(alias: str) -> str:
            for name, field in model.model_fields.items():
                if field.alias == alias:
                    return name
            return alias

        def get_field_name(loc: tuple) -> str:
            if not loc:
                return "unknown"
            return alias_to_name(loc[0]) if isinstance(loc[0], str) else str(loc[0])

        error_messages = "\n".join(
            f" - {get_field_name(e['loc'])}: {e['msg']}" for e in errors
        )
        message = (
            f"{stage} validation failed for endpoint `{url_path}`:\n{error_messages}"
        )
        super().__init__(message)
        self.errors = errors


class BaseResponse(TypedDict, total=False):
    rate_limit_limit: Optional[str]
    rate_limit_remaining: Optional[str]
    retry_after: Optional[str]


class HTTPClient(BaseClient):
    """HTTP client for making API requests.

    Args:
        config (Union[Dict[str, Any], HTTPConfig]): Client configuration.
    """

    def __init__(self, config: Union[Dict[str, Any], HTTPConfig]):
        super().__init__(config)
        self.config = HTTPConfig.model_validate(config)
        self.base_url = self.config.base_url
        self.timeout = self.config.timeout
        self.session = self._setup_session()
        self.rate_limit_headers = self.config.rate_limit_headers

    def _setup_session(self):
        """Sets up an HTTP session.

        Returns:
            requests.Session: Configured session object.
        """
        return requests.Session()

    def _handle_exception(self, response):
        """Handles HTTP exceptions.

        Args:
            response (Response): The HTTP response object.

        Raises:
            HTTPError: If response indicates an error occurred.
        """
        http_error_msg = ""
        reason = response.reason

        if 400 <= response.status_code < 500:
            if (
                response.status_code == 403
                and "'error_details':'Missing required scopes'" in response.text
            ):
                http_error_msg = f"{response.status_code} Client Error: Missing Required Scopes. Please verify your API keys include the necessary permissions."
            else:
                http_error_msg = (
                    f"{response.status_code} Client Error: {reason} {response.text}"
                )
        elif 500 <= response.status_code < 600:
            http_error_msg = (
                f"{response.status_code} Server Error: {reason} {response.text}"
            )

        if http_error_msg:
            self.logger.error(f"HTTP Error: {http_error_msg}")
            raise HTTPError(http_error_msg, response=response)

    def get(self, url_path, params: Optional[dict] = None, **kwargs) -> Dict[str, Any]:
        """Sends a GET request.

        Args:
            url_path (str): The URL path. Required.
            params (dict, optional): The query parameters. Optional.
            **kwargs: Additional arguments to pass to the request. Optional.

        Returns:
            Dict[str, Any]: The response data.
        """
        params = params or {}

        if kwargs:
            params.update(kwargs)

        return self.prepare_and_send_request("GET", url_path, params, data=None)

    def get_validated(
        self,
        url_path,
        request_model: Type[BaseModel],
        response_model: Type[BaseModel],
        **kwargs,
    ) -> BaseModel:
        """Sends a GET request including type validation of both the input and output from provided models.

        Args:
            url_path (str): The URL path. Required.
            request_model (Type[BaseModel]): Pydantic model for request validation. Required.
            response_model (Type[BaseModel]): Pydantic model for response validation. Required.
            **kwargs: Includes all arguments to pass to the request. Optional.

        Returns:
            BaseModel: The response data, validated against the response_model.
        """
        try:
            validated_params = request_model.model_validate(kwargs, by_name=True)
        except ValidationError as e:
            raise ValidationException(
                "Request", url_path, request_model, e.errors()
            ) from None
        params = validated_params.model_dump(
            mode="json", exclude_none=True, by_alias=True
        )

        response_data = self.prepare_and_send_request(
            "GET", url_path, params, data=None
        )
        validated_response = response_model.model_validate(response_data)
        return validated_response

    def post(
        self,
        url_path,
        params: Optional[dict] = None,
        data: Optional[dict] = None,
        **kwargs,
    ) -> Dict[str, Any]:
        """Sends a POST request.

        Args:
            url_path (str): The URL path. Required.
            params (dict, optional): The query parameters. Optional.
            data (dict, optional): The request body. Optional.
            **kwargs: Additional arguments to pass to the request. Optional.

        Returns:
            Dict[str, Any]: The response data.
        """
        data = data or {}

        if kwargs:
            data.update(kwargs)

        return self.prepare_and_send_request("POST", url_path, params, data)

    def put(
        self,
        url_path,
        params: Optional[dict] = None,
        data: Optional[dict] = None,
        **kwargs,
    ) -> Dict[str, Any]:
        """Sends a PUT request.

        Args:
            url_path (str): The URL path. Required.
            params (dict, optional): The query parameters. Optional.
            data (dict, optional): The request body. Optional.
            **kwargs: Additional arguments to pass to the request. Optional.

        Returns:
            Dict[str, Any]: The response data.
        """
        data = data or {}

        if kwargs:
            data.update(kwargs)

        return self.prepare_and_send_request("PUT", url_path, params, data)

    def delete(
        self,
        url_path,
        params: Optional[dict] = None,
        data: Optional[dict] = None,
        **kwargs,
    ) -> Dict[str, Any]:
        """Sends a DELETE request.

        Args:
            url_path (str): The URL path. Required.
            params (dict, optional): The query parameters. Optional.
            data (dict, optional): The request body. Optional.
            **kwargs: Additional arguments to pass to the request. Optional.

        Returns:
            Dict[str, Any]: The response data.
        """
        data = data or {}

        if kwargs:
            data.update(kwargs)

        return self.prepare_and_send_request("DELETE", url_path, params, data)

    def prepare_and_send_request(
        self,
        http_method,
        url_path,
        params: Optional[dict] = None,
        data: Optional[dict] = None,
    ):
        """Prepares and sends an HTTP request.

        Args:
            http_method (str): The HTTP method. Required.
            url_path (str): The URL path. Required.
            params (dict, optional): The query parameters. Optional.
            data (dict, optional): The request body. Optional.

        Returns:
            Dict[str, Any]: The response data.
        """
        headers = self.set_headers(http_method, url_path)

        if params is not None:
            params = {
                key: str(value).lower() if isinstance(value, bool) else value
                for key, value in params.items()
                if value is not None
            }

        if data is not None:
            data = {key: value for key, value in data.items() if value is not None}

        return self.send_request(http_method, url_path, params, headers, data=data)

    def send_request(self, http_method, url_path, params, headers, data=None):
        """Sends an HTTP request.

        Args:
            http_method (str): The HTTP method. Required.
            url_path (str): The URL path. Required.
            params (dict): The query parameters. Required.
            headers (dict): The request headers. Required.
            data (dict, optional): The request body. Optional.

        Returns:
            Dict[str, Any]: The response data.

        Raises:
            HTTPError: If the request fails.
        """
        if data is None:
            data = {}

        url = f"{self.base_url}{url_path}"

        self.logger.debug(f"Sending {http_method} request to {url}")

        response = self.session.request(
            http_method,
            url,
            params=params,
            json=data,
            headers=headers,
            timeout=self.timeout,
        )
        self._handle_exception(response)  # Raise an HTTPError for bad responses

        self.logger.debug(f"Raw response: {response.json()}")

        response_data = response.json()

        if self.rate_limit_headers:
            response_headers = dict(response.headers)
            specific_headers = {
                REST_COMMON_FIELDS.get(key, key): response_headers.get(key, None)
                for key in RATE_LIMIT_HEADERS
            }

            response_data = {**response_data, **specific_headers}

        return response_data

    def set_headers(self, method, path):
        """Sets the request headers.

        Args:
            method (str): The HTTP method. Required.
            path (str): The URL path. Required.

        Returns:
            dict: The request headers.
        """

        return {
            "User-Agent": USER_AGENT,
            "Content-Type": "application/json",
        }



"""Pure async tests for REST info endpoints that require an account."""

import pytest
from typing import List
from pydantic import BaseModel
from decimal import Decimal


@pytest.mark.asyncio
async def test_rest_info_subaccount_balances(async_rc, network):
    sid = (await async_rc.subaccounts())[0].id
    balances = await async_rc.get_subaccount_balances(subaccount_id=sid)
    assert isinstance(balances, List)
    assert all(isinstance(sb, BaseModel) for sb in balances)


@pytest.mark.asyncio
async def test_rest_info_orders(async_rc, network):
    sid = (await async_rc.subaccounts())[0].id
    orders = await async_rc.list_orders(subaccount_id=sid)
    assert isinstance(orders, List)
    assert all(isinstance(o, BaseModel) for o in orders)


@pytest.mark.asyncio
async def test_rest_info_fills(async_rc, network):
    sid = (await async_rc.subaccounts())[0].id
    fills = await async_rc.list_fills(subaccount_id=sid)
    assert isinstance(fills, List)
    assert all(isinstance(f, BaseModel) for f in fills)


@pytest.mark.asyncio
async def test_rest_info_fills_paginated(async_rc, network):
    sid = (await async_rc.subaccounts())[0].id
    fills = await async_rc._get_pages(
        endpoint="order/fill",
        request_model=async_rc._models.V1OrderFillGetParametersQuery,
        response_model=async_rc._models.PageOfOrderFillDtos,
        subaccount_id=sid,
        limit=200,
        paginate=True,
    )
    assert isinstance(fills, List)
    assert all(isinstance(f, BaseModel) for f in fills)


@pytest.mark.asyncio
async def test_rest_info_trades(async_rc, network):
    products = await async_rc.list_products()
    params = {"product_id": products[0].id, "order": "desc", "limit": 100}
    trades = await async_rc.list_trades(**params)
    assert isinstance(trades, List)
    assert all(isinstance(t, BaseModel) for t in trades)


@pytest.mark.asyncio
async def test_rest_info_positions(async_rc, network):
    sid = (await async_rc.subaccounts())[0].id
    positions = await async_rc.list_positions(subaccount_id=sid)
    assert isinstance(positions, List)
    assert all(isinstance(p, BaseModel) for p in positions)


@pytest.mark.asyncio
async def test_rest_info_maintenance_margin(async_rc, network):
    sid = (await async_rc.subaccounts())[0].id
    mm = await async_rc.get_maintenance_margin(subaccount_id=sid)
    assert isinstance(mm, Decimal)

    # specify product id
    products_by_ticker = await async_rc.products_by_ticker()
    pid = products_by_ticker["BTCUSD"].id
    btc_mm = await async_rc.get_maintenance_margin(subaccount_id=sid, product_ids=[pid])
    assert isinstance(btc_mm, Decimal)

    # specify products
    products = [
        products_by_ticker["BTCUSD"],
        products_by_ticker["ETHUSD"],
    ]
    eth_btc_mm = await async_rc.get_maintenance_margin(
        subaccount_id=sid, products=products
    )
    assert isinstance(eth_btc_mm, Decimal)

    # specify product dicts
    product_dicts = [
        {
            "id": products_by_ticker["BTCUSD"].id,
            "max_leverage": 5,
            "taker_fee": "0.001",
        },
        {
            "id": products_by_ticker["ETHUSD"].id,
            "max_leverage": 5,
            "taker_fee": "0.001",
        },
    ]
    eth_btc_mm = await async_rc.get_maintenance_margin(
        subaccount_id=sid, products=product_dicts
    )
    assert isinstance(eth_btc_mm, Decimal)



import time
from typing import List, Optional, Union
from decimal import Decimal
from uuid import UUID
from ethereal.constants import API_PREFIX
from ethereal.rest.util import (
    generate_nonce,
    uuid_to_bytes32,
    client_order_id_to_bytes32,
)
from ethereal.models.rest import (
    OrderDto,
    OrderFillDto,
    TradeDto,
    SubmitOrderDto,
    SubmitOrderLimitDtoData,
    SubmitOrderMarketDtoData,
    CancelOrderDto,
    CancelOrderResultDto,
    DryRunOrderCreatedDto,
    SubmitOrderCreatedDto,
)


async def list_orders(self, **kwargs) -> List[OrderDto]:
    """Lists orders for a subaccount.

    Args:
        subaccount_id (str): UUID of the subaccount. Required.

    Other Parameters:
        product_ids (List[str], optional): Filter by one or more product IDs. Optional.
        client_order_id (str, optional): Filter by a client-generated order id. Optional.
        statuses (List[str], optional): Filter by status values. Optional.
        created_after (float, optional): Filter orders created after this timestamp. Optional.
        created_before (float, optional): Filter orders created before this timestamp. Optional.
        side (int, optional): Filter by order side (0 for buy, 1 for sell). Optional.
        close (bool, optional): Filter by close flag. Optional.
        stop_types (List[int], optional): Filter by stop types (0 for GAIN, 1 for LOSS). Optional.
        is_working (bool, optional): Filter working orders. Optional.
        is_pending (bool, optional): Filter pending orders. Optional.
        order (str, optional): Sort order, 'asc' or 'desc'. Optional.
        limit (float, optional): Maximum number of results to return. Optional.
        cursor (str, optional): Pagination cursor for fetching the next page. Optional.
        order_by (str, optional): Field to order by (e.g., 'createdAt'). Optional.
        **kwargs: Additional request parameters accepted by the API. Optional.

    Returns:
        List[OrderDto]: Order records for the subaccount.
    """
    res = await self.get_validated(
        url_path=f"{API_PREFIX}/order",
        request_model=self._models.V1OrderGetParametersQuery,
        response_model=self._models.PageOfOrderDtos,
        **kwargs,
    )
    data = [
        self._models.OrderDto(**model.model_dump(by_alias=True)) for model in res.data
    ]
    return data


async def list_fills(self, **kwargs) -> List[OrderFillDto]:
    """Lists order fills for a subaccount.

    Args:
        subaccount_id (str): UUID of the subaccount. Required.

    Other Parameters:
        product_ids (List[str], optional): Filter by one or more product IDs. Optional.
        created_after (float, optional): Filter fills created after this timestamp. Optional.
        created_before (float, optional): Filter fills created before this timestamp. Optional.
        side (int, optional): Filter by order side (0 for buy, 1 for sell). Optional.
        include_self_trades (bool, optional): Include self trades in results. Defaults to False. Optional.
        order (str, optional): Sort order, 'asc' or 'desc'. Optional.
        limit (float, optional): Maximum number of results to return. Optional.
        cursor (str, optional): Pagination cursor for fetching the next page. Optional.
        order_by (str, optional): Field to order by (e.g., 'createdAt', 'productId'). Optional.
        **kwargs: Additional request parameters accepted by the API. Optional.

    Returns:
        List[OrderFillDto]: Fill records for the subaccount.
    """
    res = await self.get_validated(
        url_path=f"{API_PREFIX}/order/fill",
        request_model=self._models.V1OrderFillGetParametersQuery,
        response_model=self._models.PageOfOrderFillDtos,
        **kwargs,
    )
    data = [
        self._models.OrderFillDto(**model.model_dump(by_alias=True))
        for model in res.data
    ]
    return data


async def list_trades(self, **kwargs) -> List[TradeDto]:
    """Lists trades for a specific product.

    Args:
        product_id (str): Product ID to query trades for. Required.

    Other Parameters:
        order (str, optional): Sort order, 'asc' or 'desc'. Optional.
        limit (float, optional): Maximum number of results to return. Optional.
        cursor (str, optional): Pagination cursor for fetching the next page. Optional.
        order_by (str, optional): Field to order by (e.g., 'createdAt'). Optional.
        **kwargs: Additional request parameters accepted by the API. Optional.

    Returns:
        List[TradeDto]: Trade records.
    """
    res = await self.get_validated(
        url_path=f"{API_PREFIX}/order/trade",
        request_model=self._models.V1OrderTradeGetParametersQuery,
        response_model=self._models.PageOfTradeDtos,
        **kwargs,
    )
    data = [
        self._models.TradeDto(**model.model_dump(by_alias=True)) for model in res.data
    ]
    return data


async def get_order(self, id: UUID, **kwargs) -> OrderDto:
    """Gets a specific order by ID.

    Args:
        id (str): UUID of the order. Required.

    Other Parameters:
        **kwargs: Additional request parameters accepted by the API. Optional.

    Returns:
        OrderDto: Order details.
    """
    endpoint = f"{API_PREFIX}/order/{id}"
    response = await self.get(endpoint, **kwargs)
    return self._models.OrderDto(**response)


async def prepare_order(
    self,
    sender: str,
    price: Optional[Union[str, float, Decimal]] = None,
    quantity: Optional[Union[str, float, Decimal]] = None,
    side: Optional[int] = None,
    subaccount: Optional[str] = None,
    onchain_id: Optional[float] = None,
    order_type: Optional[str] = None,
    client_order_id: Optional[str] = None,
    time_in_force: Optional[str] = None,
    post_only: Optional[bool] = False,
    reduce_only: Optional[bool] = False,
    close: Optional[bool] = None,
    stop_price: Optional[Union[str, float, Decimal]] = None,
    stop_type: Optional[int] = None,
    group_id: Optional[str] = None,
    group_contingency_type: Optional[int] = None,
    expires_at: Optional[int] = None,
    include_signature: bool = False,
    **kwargs,
) -> SubmitOrderDto:
    """Prepares the payload for an order, optionally including a signature.

    Args:
        sender (str): Address placing the order. Required.
        price (Union[str, float, Decimal], optional): Limit price for LIMIT orders.
        quantity (Union[str, float, Decimal], optional): Order size.
        side (int, optional): 0 for buy, 1 for sell.
        subaccount (str, optional): Hex-encoded subaccount name.
        onchain_id (float, optional): Product onchain ID.
        order_type (str, optional): 'LIMIT' or 'MARKET'.
        client_order_id (str, optional): Subaccount-scoped client-generated id (UUID or <=32 alphanumeric).
        time_in_force (str, optional): For LIMIT orders (e.g., 'GTC', 'GTD').
        post_only (bool, optional): For LIMIT orders; rejects if crossing.
        reduce_only (bool, optional): If True, order only reduces position.
        close (bool, optional): If True, closes the position.
        stop_price (Union[str, float, Decimal], optional): Stop trigger price.
        stop_type (int, optional): Stop type enum value.
        group_id (str, optional): Contingency group id.
        group_contingency_type (int, optional): Group contingency type.
        expires_at (int, optional): Expiry timestamp for GTD.
        include_signature (bool): If True, sign the payload immediately.

    Other Parameters:
        nonce (str, optional): Custom nonce for signing.
        signed_at (int, optional): Seconds since epoch for signature timestamp.
        **kwargs: Additional request parameters accepted by the API.

    Returns:
        SubmitOrderDto: Prepared (and optionally signed) order payload.
    """
    # Generate nonce and signed_at timestamp
    nonce = kwargs.get("nonce", None) or generate_nonce()
    signed_at = kwargs.get("signed_at", None) or int(time.time())

    # Prepare order data with common fields
    order_data = {
        "sender": sender,
        "subaccount": subaccount,
        "quantity": quantity,
        "price": price,
        "side": side,
        "engineType": 0,
        "onchainId": onchain_id,
        "nonce": nonce,
        "type": order_type,
        "clientOrderId": client_order_id,
        "reduceOnly": reduce_only,
        "signedAt": signed_at,
        "close": close,
        "stopPrice": stop_price,
        "stopType": stop_type,
        "groupId": group_id,
        "groupContingencyType": group_contingency_type,
        "expiresAt": expires_at,
    }

    # Declare data_model type
    data_model: Union[SubmitOrderLimitDtoData, SubmitOrderMarketDtoData]

    # Create specific order data based on type
    if order_type == "LIMIT":
        order_data.update(
            {
                "timeInForce": time_in_force,
                "postOnly": post_only,
            }
        )
        data_model = self._models.SubmitOrderLimitDtoData.model_validate(order_data)
    elif order_type == "MARKET":
        data_model = self._models.SubmitOrderMarketDtoData.model_validate(order_data)
    else:
        raise ValueError(f"Invalid order type: {order_type}")

    result = self._models.SubmitOrderDto.model_validate(
        {
            "data": data_model.model_dump(
                mode="json", by_alias=True, exclude_unset=True, exclude_none=True
            ),
            "signature": "",
        }
    )

    if include_signature:
        result = await self.sign_order(result)

    return result


async def sign_order(
    self, order: SubmitOrderDto, private_key: Optional[str] = None
) -> SubmitOrderDto:
    """Signs an order payload using EIP-712.

    Args:
        order (SubmitOrderDto): Prepared order to sign. Required.
        private_key (str, optional): Private key override. Defaults to client's key.

    Returns:
        SubmitOrderDto: The same DTO with signature populated.

    Raises:
        ValueError: If no chain client or private key is available.
    """
    if not hasattr(self, "chain") or not self.chain:
        raise ValueError("No chain client available for signing")
    if not private_key and not self.chain.private_key:
        raise ValueError("No private key available for signing")
    elif not private_key:
        private_key = self.chain.private_key

    # Update message signedAt
    order.data.signed_at = int(time.time())

    # Prepare message for signing
    message = order.data.model_dump(mode="json", by_alias=True)

    # Make some adjustments to the message
    message["quantity"] = int(Decimal(message["quantity"]) * Decimal("1e9"))
    message["price"] = int(Decimal(message.get("price", 0)) * Decimal("1e9"))
    message["productId"] = int(message["onchainId"])
    message["signedAt"] = int(message["signedAt"])
    if message.get("clientOrderId"):
        message["clientOrderId"] = client_order_id_to_bytes32(message["clientOrderId"])

    # Get domain and types for signing
    primary_type = "TradeOrder"
    domain = self.rpc_config.domain.model_dump(mode="json", by_alias=True)
    types = self.chain.get_signature_types(self.rpc_config, primary_type)

    # Sign the message
    order.signature = self.chain.sign_message(
        self.chain.private_key, domain, types, primary_type, message
    )
    return order


async def submit_order(
    self,
    order: SubmitOrderDto,
    **kwargs,
) -> SubmitOrderCreatedDto:
    """Submits a prepared and signed order.

    Args:
        order (SubmitOrderDto): Prepared and signed order payload. Required.

    Other Parameters:
        **kwargs: Additional request parameters accepted by the API. Optional.

    Returns:
        SubmitOrderCreatedDto: Created order response.
    """
    endpoint = f"{API_PREFIX}/order"
    res = await self.post(
        endpoint,
        data=order.model_dump(
            mode="json", by_alias=True, exclude_unset=True, exclude_none=True
        ),
        **kwargs,
    )
    return self._models.SubmitOrderCreatedDto.model_validate(res)


async def dry_run_order(
    self,
    order: SubmitOrderDto,
    **kwargs,
) -> DryRunOrderCreatedDto:
    """Submits a prepared order for validation without execution.

    Args:
        order (SubmitOrderDto): Prepared order payload. Required.

    Other Parameters:
        **kwargs: Additional request parameters accepted by the API. Optional.

    Returns:
        DryRunOrderCreatedDto: Dry-run validation result.
    """
    submit_payload = self._models.SubmitDryOrderDto.model_validate(
        {"data": order.data.model_dump(mode="json", by_alias=True, exclude_unset=True)}
    )
    endpoint = f"{API_PREFIX}/order/dry-run"
    res = await self.post(
        endpoint,
        data=submit_payload.model_dump(
            mode="json", by_alias=True, exclude_unset=True, exclude_none=True
        ),
        **kwargs,
    )
    return self._models.DryRunOrderCreatedDto.model_validate(res)


async def prepare_cancel_order(
    self,
    sender: str,
    subaccount: str,
    order_ids: List[UUID] = [],
    client_order_ids: List[str] = [],
    include_signature: bool = False,
    **kwargs,
) -> CancelOrderDto:
    """Prepares the payload for canceling one or more orders.

    Args:
        sender (str): Address initiating the cancellation. Required.
        subaccount (str): Hex-encoded subaccount name. Required.
        order_ids (List[str]): Order UUIDs to cancel. Optional.
        client_order_ids (List[str]): Client-generated IDs to cancel. Optional.
        include_signature (bool): If True, sign the payload immediately. Optional.

    Other Parameters:
        nonce (str, optional): Custom nonce for signing.
        **kwargs: Additional request parameters accepted by the API.

    Returns:
        CancelOrderDto: Prepared (and optionally signed) cancel payload.
    """
    nonce = kwargs.get("nonce", None) or generate_nonce()
    uuid_order_ids = [
        UUID(order_id) if isinstance(order_id, str) else order_id
        for order_id in order_ids
    ]
    data_model = self._models.CancelOrderDtoData(
        sender=sender,
        subaccount=subaccount,
        nonce=nonce,
        orderIds=uuid_order_ids,
        clientOrderIds=client_order_ids,
    )
    result = self._models.CancelOrderDto.model_validate(
        {"data": data_model.model_dump(mode="json", by_alias=True), "signature": ""}
    )
    if include_signature:
        return await self.sign_cancel_order(result)
    return result


async def sign_cancel_order(
    self,
    order_to_cancel: CancelOrderDto,
    private_key: Optional[str] = None,
) -> CancelOrderDto:
    """Signs a cancel order payload using EIP-712.

    Args:
        order_to_cancel (CancelOrderDto): Prepared cancel payload. Required.
        private_key (str, optional): Private key override. Defaults to client's key.

    Returns:
        CancelOrderDto: The same DTO with signature populated.

    Raises:
        ValueError: If no chain client or private key is available.
    """
    if not hasattr(self, "chain") or not self.chain:
        raise ValueError("No chain client available for signing")
    if not private_key and not self.chain.private_key:
        raise ValueError("No private key available for signing")
    elif not private_key:
        private_key = self.chain.private_key

    # Prepare message for signing
    message = order_to_cancel.data.model_dump(mode="json", by_alias=True)

    # For cancel orders, orderIds need to be converted to bytes32 format
    order_ids = (
        [uuid_to_bytes32(str(order_id)) for order_id in order_to_cancel.data.order_ids]
        if order_to_cancel.data.order_ids
        else []
    )
    message["orderIds"] = order_ids

    client_order_ids = (
        [client_order_id_to_bytes32(id) for id in order_to_cancel.data.client_order_ids]
        if order_to_cancel.data.client_order_ids
        else []
    )
    message["clientOrderIds"] = client_order_ids

    # Get domain and types for signing
    primary_type = "CancelOrder"
    domain = self.rpc_config.domain.model_dump(mode="json", by_alias=True)
    types = self.chain.get_signature_types(self.rpc_config, primary_type)

    # Sign the message
    order_to_cancel.signature = self.chain.sign_message(
        self.chain.private_key, domain, types, primary_type, message
    )
    return order_to_cancel


async def cancel_order(
    self,
    order_to_cancel: CancelOrderDto,
    **kwargs,
) -> List[CancelOrderResultDto]:
    """Submits a prepared and signed cancel order request.

    Args:
        order_to_cancel (CancelOrderDto): Prepared and signed cancel payload. Required.

    Other Parameters:
        **kwargs: Additional request parameters accepted by the API. Optional.

    Returns:
        List[CancelOrderResultDto]: Cancellation results per order id.
    """
    endpoint = f"{API_PREFIX}/order/cancel"
    res = await self.post(
        endpoint,
        data=order_to_cancel.model_dump(mode="json", by_alias=True, exclude_none=True),
        **kwargs,
    )
    return [
        self._models.CancelOrderResultDto.model_validate(item)
        for item in res.get("data", [])
    ]



from web3 import Web3
from ethereal.models.config import (
    BaseConfig,
    WSConfig,
    ChainConfig,
    RESTConfig,
)
from ethereal.base_client import BaseClient
from ethereal.ws.ws_base import WSBase
from ethereal.chain_client import ChainClient
from ethereal.rest_client import RESTClient


def test_base_client_with_dict():
    """Test BaseClient initialization with dict config."""
    bc = BaseClient({"verbose": True})
    assert bc is not None


def test_base_client_with_class():
    """Test BaseClient initialization with BaseConfig class."""
    config = BaseConfig(verbose=True)
    bc = BaseClient(config)
    assert bc is not None


def test_ws_client_with_dict(network_config):
    """Test WSBase initialization with dict config using network fixture."""
    wsc = WSBase({"base_url": network_config["ws_url"], "verbose": True})
    assert wsc is not None
    assert str(wsc.base_url) == network_config["ws_url"]


def test_ws_client_with_class(network_config):
    """Test WSBase initialization with WSConfig class using network fixture."""
    config = WSConfig(base_url=network_config["ws_url"], verbose=True)
    wsc = WSBase(config)
    assert wsc is not None
    assert str(wsc.base_url) == network_config["ws_url"]


def test_chain_client_with_dict(network_config):
    """Test ChainClient initialization with dict config using network fixture."""
    test_account = Web3().eth.account.create()
    private_key = test_account.key.hex()

    cc = ChainClient(
        {
            "rpc_url": network_config["rpc_url"],
            "private_key": private_key,
        }
    )
    assert cc is not None
    assert cc.chain_id == network_config["chain_id"]


def test_chain_client_with_class(network_config):
    """Test ChainClient initialization with ChainConfig class using network fixture."""
    test_account = Web3().eth.account.create()
    private_key = test_account.key.hex()

    config = ChainConfig(
        rpc_url=network_config["rpc_url"],
        private_key=private_key,
    )
    cc = ChainClient(config)
    assert cc is not None
    assert cc.chain_id == network_config["chain_id"]


def test_rest_client_with_dict(network_config):
    """Test RESTClient initialization with dict config using network fixture."""
    test_account = Web3().eth.account.create()
    private_key = test_account.key.hex()

    rc = RESTClient(
        {
            "base_url": network_config["base_url"],
            "chain_config": {
                "private_key": private_key,
                "rpc_url": network_config["rpc_url"],
            },
        }
    )
    assert rc is not None
    assert rc._async_client.chain.chain_id == network_config["chain_id"]
    rc.close()


def test_rest_client_with_class(network_config):
    """Test RESTClient initialization with RESTConfig class using network fixture."""
    test_account = Web3().eth.account.create()
    private_key = test_account.key.hex()

    chain_config = ChainConfig(
        rpc_url=network_config["rpc_url"],
        private_key=private_key,
    )

    config = RESTConfig(
        base_url=network_config["base_url"],
        chain_config=chain_config,
    )
    rc = RESTClient(config)
    assert rc is not None
    assert rc._async_client.chain.chain_id == network_config["chain_id"]
    rc.close()


def test_rest_client_without_chain():
    """Test RESTClient initialization without chain config."""
    rc = RESTClient()
    assert rc is not None
    assert rc._async_client.chain is None
    rc.close()  # Properly cleanup


def test_rest_client_with_empty_config():
    """Test RESTClient initialization with empty RESTConfig."""
    config = RESTConfig()
    rc = RESTClient(config)
    assert rc is not None
    assert rc._async_client.chain is None
    rc.close()  # Properly cleanup


def test_rest_client_with_context_manager():
    """Test RESTClient as context manager for automatic cleanup."""
    with RESTClient() as rc:
        assert rc is not None
        assert rc._async_client.chain is None
    # Client should be closed automatically


def test_sync_client_list_products():
    """Test sync client listing products."""
    rc = RESTClient()
    try:
        products = rc.list_products()
        assert products is not None
        assert isinstance(products, list)
    finally:
        rc.close()


def test_client_with_network_parameter(network, network_config):
    """Test client initialization with network parameter."""
    test_account = Web3().eth.account.create()
    private_key = test_account.key.hex()

    rc = RESTClient(
        {
            "network": network,
            "base_url": network_config["base_url"],
            "chain_config": {
                "chain_id": network_config["chain_id"],
                "rpc_url": network_config["rpc_url"],
                "private_key": private_key,
            },
        }
    )
    assert rc is not None
    assert rc._async_client.chain.chain_id == network_config["chain_id"]

    if hasattr(rc._async_client, "config") and hasattr(
        rc._async_client.config, "network"
    ):
        assert rc._async_client.config.network == network

    rc.close()


def test_read_only_client(network_config):
    """Test read-only client initialization (address instead of private_key)."""
    test_account = Web3().eth.account.create()
    address = test_account.address

    rc = RESTClient(
        {
            "base_url": network_config["base_url"],
            "chain_config": {
                "chain_id": network_config["chain_id"],
                "rpc_url": network_config["rpc_url"],
                "address": address,
            },
        }
    )
    assert rc is not None
    assert rc._async_client.chain.chain_id == network_config["chain_id"]
    assert rc._async_client.chain.address == address
    rc.close()


def test_verbose_logging_configs(network_config):
    """Test verbose logging configuration across different clients."""
    verbose = True

    ws_config = WSConfig(base_url=network_config["ws_url"], verbose=verbose)
    wsc = WSBase(ws_config)
    assert wsc is not None
    assert wsc.config.verbose == verbose



from typing import Any, Dict, Optional, Union, TypedDict, Type
from pydantic import BaseModel, ValidationError
import logging

import httpx

from ethereal.base_client import BaseClient
from ethereal.models.config import HTTPConfig
from ethereal.constants import RATE_LIMIT_HEADERS, USER_AGENT, REST_COMMON_FIELDS


class ValidationException(Exception):
    def __init__(self, stage: str, url_path: str, model: Type[BaseModel], errors: list):
        def alias_to_name(alias: str) -> str:
            for name, field in model.model_fields.items():
                if field.alias == alias:
                    return name
            return alias

        def get_field_name(loc: tuple) -> str:
            if not loc:
                return "unknown"
            return alias_to_name(loc[0]) if isinstance(loc[0], str) else str(loc[0])

        error_messages = "\n".join(
            f" - {get_field_name(e['loc'])}: {e['msg']}" for e in errors
        )
        message = (
            f"{stage} validation failed for endpoint `{url_path}`:\n{error_messages}"
        )
        super().__init__(message)
        self.errors = errors


class BaseResponse(TypedDict, total=False):
    rate_limit_limit: Optional[str]
    rate_limit_remaining: Optional[str]
    retry_after: Optional[str]


class AsyncHTTPClient(BaseClient):
    """Asynchronous HTTP client for making API requests.

    Maintainer notes:
    - We use a single `httpx.AsyncClient` per instance; call `await close()`
      when done to free sockets. The higher-level REST client wires this up.
    - `get_validated` validates both request params and responses using
      pydantic models, raising `ValidationException` with human-friendly field
      names. Endpoint modules rely on this for type safety at boundaries.
    - Auth: extend `set_headers()` to inject auth headers or request signing
      when needed; keeping it centralized avoids per-endpoint duplication.
    - Future: consider passing `base_url` to `AsyncClient` and using relative
      paths to simplify URL construction. Current approach is explicit.
    """

    def __init__(self, config: Union[Dict[str, Any], HTTPConfig]):
        super().__init__(config)
        self.config = HTTPConfig.model_validate(config)
        self.base_url = self.config.base_url
        # Normalize base URL once to avoid double slashes when joining paths.
        self._base_url_str = str(self.base_url).rstrip("/") if self.base_url else ""
        self.timeout = self.config.timeout

        # Configure httpx logging - suppress INFO logs unless verbose mode is enabled
        if not getattr(self.config, "verbose", False):
            httpx_logger = logging.getLogger("httpx")
            httpx_logger.setLevel(logging.WARNING)

        self.session = httpx.AsyncClient()
        self.rate_limit_headers = self.config.rate_limit_headers

    async def close(self):
        """Closes the HTTP session."""
        if hasattr(self, "session") and not self.session.is_closed:
            await self.session.aclose()

    def _handle_exception(self, response: httpx.Response):
        """Handles HTTP exceptions."""
        http_error_msg = ""
        reason = response.reason_phrase

        if 400 <= response.status_code < 500:
            if (
                response.status_code == 403
                and "'error_details':'Missing required scopes'" in response.text
            ):
                http_error_msg = f"{response.status_code} Client Error: Missing Required Scopes. Please verify your API keys include the necessary permissions."
            else:
                http_error_msg = (
                    f"{response.status_code} Client Error: {reason} {response.text}"
                )
        elif 500 <= response.status_code < 600:
            http_error_msg = (
                f"{response.status_code} Server Error: {reason} {response.text}"
            )

        if http_error_msg:
            self.logger.error(f"HTTP Error: {http_error_msg}")
            raise httpx.HTTPStatusError(
                http_error_msg, request=response.request, response=response
            )

    async def get(
        self,
        url_path,
        params: Optional[dict] = None,
        *,
        base_url_override: Optional[str] = None,
        **kwargs,
    ) -> Dict[str, Any]:
        """Sends a GET request."""
        params = params or {}
        if kwargs:
            params.update(kwargs)
        return await self.prepare_and_send_request(
            "GET",
            url_path,
            params,
            data=None,
            base_url_override=base_url_override,
        )

    async def get_validated(
        self,
        url_path,
        request_model: Type[BaseModel],
        response_model: Type[BaseModel],
        *,
        base_url_override: Optional[str] = None,
        **kwargs,
    ) -> BaseModel:
        """Sends a GET request with type validation."""
        try:
            validated_params = request_model.model_validate(kwargs, by_name=True)
        except ValidationError as e:
            raise ValidationException(
                "Request", url_path, request_model, e.errors()
            ) from None
        params = validated_params.model_dump(
            mode="json", exclude_none=True, by_alias=True
        )

        response_data = await self.prepare_and_send_request(
            "GET",
            url_path,
            params,
            data=None,
            base_url_override=base_url_override,
        )
        validated_response = response_model.model_validate(response_data)
        return validated_response

    async def post(
        self,
        url_path,
        params: Optional[dict] = None,
        data: Optional[dict] = None,
        *,
        base_url_override: Optional[str] = None,
        **kwargs,
    ) -> Dict[str, Any]:
        """Sends a POST request."""
        data = data or {}
        if kwargs:
            data.update(kwargs)
        return await self.prepare_and_send_request(
            "POST",
            url_path,
            params,
            data,
            base_url_override=base_url_override,
        )

    async def put(
        self,
        url_path,
        params: Optional[dict] = None,
        data: Optional[dict] = None,
        *,
        base_url_override: Optional[str] = None,
        **kwargs,
    ) -> Dict[str, Any]:
        """Sends a PUT request."""
        data = data or {}
        if kwargs:
            data.update(kwargs)
        return await self.prepare_and_send_request(
            "PUT",
            url_path,
            params,
            data,
            base_url_override=base_url_override,
        )

    async def delete(
        self,
        url_path,
        params: Optional[dict] = None,
        data: Optional[dict] = None,
        *,
        base_url_override: Optional[str] = None,
        **kwargs,
    ) -> Dict[str, Any]:
        """Sends a DELETE request."""
        data = data or {}
        if kwargs:
            data.update(kwargs)
        return await self.prepare_and_send_request(
            "DELETE",
            url_path,
            params,
            data,
            base_url_override=base_url_override,
        )

    async def prepare_and_send_request(
        self,
        http_method,
        url_path,
        params: Optional[dict] = None,
        data: Optional[dict] = None,
        *,
        base_url_override: Optional[str] = None,
    ):
        """Prepares and sends an HTTP request."""
        headers = self.set_headers(http_method, url_path)

        if params is not None:
            params = {
                key: str(value).lower() if isinstance(value, bool) else value
                for key, value in params.items()
                if value is not None
            }

        if data is not None:
            data = {key: value for key, value in data.items() if value is not None}

        return await self.send_request(
            http_method,
            url_path,
            params,
            headers,
            data=data,
            base_url_override=base_url_override,
        )

    async def send_request(
        self,
        http_method,
        url_path,
        params,
        headers,
        data=None,
        *,
        base_url_override: Optional[str] = None,
    ):
        """Sends an HTTP request."""
        if data is None:
            data = {}

        # Build URL with exactly one slash between base and path.
        path = url_path if str(url_path).startswith("/") else f"/{url_path}"
        base = base_url_override or self._base_url_str
        url = f"{base}{path}"
        self.logger.debug(f"Sending {http_method} request to {url}")

        response = await self.session.request(
            http_method,
            url,
            params=params,
            json=data,
            headers=headers,
            timeout=self.timeout,
        )
        self._handle_exception(response)

        self.logger.debug(f"Raw response: {response.json()}")
        response_data = response.json()

        if self.rate_limit_headers:
            response_headers = dict(response.headers)
            specific_headers = {
                REST_COMMON_FIELDS.get(key, key): response_headers.get(key, None)
                for key in RATE_LIMIT_HEADERS
            }
            response_data = {**response_data, **specific_headers}

        return response_data

    def set_headers(self, method, path):
        """Sets the request headers.

        This is the extension point for auth. If the API introduces API keys or
        request signing, add them here to keep all endpoints consistent.
        """
        return {
            "User-Agent": USER_AGENT,
            "Content-Type": "application/json",
        }



"""Pure async tests for REST position liquidation endpoints."""

import pytest
from typing import List
from pydantic import BaseModel


@pytest.mark.asyncio
async def test_rest_list_position_liquidations(async_rc):
    """Test listing position liquidations."""
    liquidations = await async_rc.list_position_liquidations()
    assert isinstance(liquidations, List)
    assert all(isinstance(liq, BaseModel) for liq in liquidations)



"""Pure sync tests for token operations."""

import pytest

from ethereal.rest.util import ensure_bytes32_hex

DEFAULT_DESTINATION_ENDPOINT = 0


def test_prepare_withdraw_token(rc_ro, network):
    subaccounts = rc_ro.list_subaccounts(sender=rc_ro.chain.address)
    sub = subaccounts[0]
    tokens = rc_ro.list_tokens()
    if not tokens:
        pytest.skip("No tokens available for testing")
    token = tokens[0]
    destination_endpoint = DEFAULT_DESTINATION_ENDPOINT
    destination_address = rc_ro.chain.address
    dto = rc_ro.prepare_withdraw_token(
        subaccount=sub.name,
        token=token.address,
        amount=100000,
        account=rc_ro.chain.address,
        destination_address=destination_address,
        destination_endpoint=destination_endpoint,
    )
    assert (
        isinstance(dto, rc_ro._models.InitiateWithdrawDto)
        and dto.data.token == token.address
        and dto.data.subaccount == sub.name
        and dto.data.amount == 100000
        and dto.data.account == rc_ro.chain.address
        and dto.signature == ""
        and dto.data.lz_destination_address == ensure_bytes32_hex(destination_address)
        and dto.data.lz_destination_eid.value == destination_endpoint
    )


def test_prepare_and_sign_withdraw_token(rc, network):
    subaccounts = rc.list_subaccounts(sender=rc.chain.address)
    sub = subaccounts[0]
    token = rc.list_tokens()[0]
    destination_address = rc.chain.address
    dto = rc.prepare_withdraw_token(
        subaccount=sub.name,
        token=token.address,
        amount=100000,
        account=rc.chain.address,
        destination_address=destination_address,
        destination_endpoint=DEFAULT_DESTINATION_ENDPOINT,
    )
    signed = rc.sign_withdraw_token(dto)
    assert isinstance(signed, rc._models.InitiateWithdrawDto) and signed.signature != ""


def test_prepare_with_automatic_signing(rc, network):
    subaccounts = rc.list_subaccounts(sender=rc.chain.address)
    sub = subaccounts[0]
    token = rc.list_tokens()[0]
    destination_address = rc.chain.address
    dto = rc.prepare_withdraw_token(
        subaccount=sub.name,
        token=token.address,
        amount=100000,
        account=rc.chain.address,
        include_signature=True,
        destination_address=destination_address,
        destination_endpoint=DEFAULT_DESTINATION_ENDPOINT,
    )
    assert isinstance(dto, rc._models.InitiateWithdrawDto) and dto.signature != ""


def test_prepare_withdraw_token_with_custom_nonce(rc_ro, network):
    subaccounts = rc_ro.list_subaccounts(sender=rc_ro.chain.address)
    sub = subaccounts[0]
    tokens = rc_ro.list_tokens()
    if not tokens:
        pytest.skip("No tokens available for testing")
    token = tokens[0]
    nonce = "123456789"
    destination_address = rc_ro.chain.address
    dto = rc_ro.prepare_withdraw_token(
        subaccount=sub.name,
        token=token.address,
        amount=100000,
        account=rc_ro.chain.address,
        destination_address=destination_address,
        destination_endpoint=DEFAULT_DESTINATION_ENDPOINT,
        nonce=nonce,
    )
    assert (
        isinstance(dto, rc_ro._models.InitiateWithdrawDto) and dto.data.nonce == nonce
    )


def test_prepare_withdraw_token_with_custom_signed_at(rc_ro, network):
    subaccounts = rc_ro.list_subaccounts(sender=rc_ro.chain.address)
    sub = subaccounts[0]
    token = rc_ro.list_tokens()[0]
    ts = 1620000000
    destination_address = rc_ro.chain.address
    dto = rc_ro.prepare_withdraw_token(
        subaccount=sub.name,
        token=token.address,
        amount=100000,
        account=rc_ro.chain.address,
        destination_address=destination_address,
        destination_endpoint=DEFAULT_DESTINATION_ENDPOINT,
        signed_at=ts,
    )
    assert (
        isinstance(dto, rc_ro._models.InitiateWithdrawDto) and dto.data.signed_at == ts
    )


def test_prepare_withdraw_token_with_custom_destination(rc_ro, network):
    subaccounts = rc_ro.list_subaccounts(sender=rc_ro.chain.address)
    sub = subaccounts[0]
    tokens = rc_ro.list_tokens()
    if not tokens:
        pytest.skip("No tokens available for testing")
    token = tokens[0]
    destination_address = rc_ro.chain.address
    destination_endpoint = 40422
    dto = rc_ro.prepare_withdraw_token(
        subaccount=sub.name,
        token=token.address,
        amount=100000,
        account=rc_ro.chain.address,
        destination_address=destination_address,
        destination_endpoint=destination_endpoint,
    )
    assert (
        isinstance(dto, rc_ro._models.InitiateWithdrawDto)
        and dto.data.token == token.address
        and dto.data.subaccount == sub.name
        and dto.data.amount == 100000
        and dto.data.account == rc_ro.chain.address
        and dto.signature == ""
        and dto.data.lz_destination_address == ensure_bytes32_hex(destination_address)
        and dto.data.lz_destination_eid.value == destination_endpoint
    )


@pytest.mark.skip(reason="This test actually submits a withdrawal request")
def test_prepare_sign_submit_withdraw_token(rc, network):
    subaccounts = rc.list_subaccounts(sender=rc.chain.address)
    sub = subaccounts[0]
    tokens = rc.list_tokens()
    token = next((t for t in tokens if t.name == "USD"))
    destination_address = rc.chain.address
    dto = rc.prepare_withdraw_token(
        subaccount=sub.name,
        token=token.address,
        amount=5,
        account=rc.chain.address,
        destination_address=destination_address,
        destination_endpoint=DEFAULT_DESTINATION_ENDPOINT,
    )
    signed = rc.sign_withdraw_token(dto)
    result = rc.withdraw_token(signed, token_id=token.id)
    assert (
        isinstance(result, rc._models.WithdrawDto)
        and result.token == token.address
        and result.subaccount == sub.name
    )


@pytest.mark.skip(reason="This test actually submits a withdrawal request")
def test_prepare_sign_submit_withdraw_token_one_step(rc, network):
    subaccounts = rc.list_subaccounts(sender=rc.chain.address)
    sub = subaccounts[0]
    tokens = rc.list_tokens()
    token = next((t for t in tokens if t.name == "USD"))
    destination_address = rc.chain.address
    dto = rc.prepare_withdraw_token(
        subaccount=sub.name,
        token=token.address,
        amount=50,
        account=rc.chain.address,
        include_signature=True,
        destination_address=destination_address,
        destination_endpoint=DEFAULT_DESTINATION_ENDPOINT,
    )
    result = rc.withdraw_token(dto, token_id=token.id)
    assert (
        isinstance(result, rc._models.WithdrawDto)
        and result.token == token.address
        and result.subaccount == sub.name
    )
